"""
defines a class for organizing and performing bulk rendering operations on
multispectral image products
"""
import gc
import logging
import os
from collections.abc import Mapping, Callable, Collection, Sequence
from pathlib import Path
from typing import Optional, Union, MutableMapping, Any

# note: ignore complaints from static analyzers about this import. dill
# performs pickling magick at import.

from cytoolz.dicttoolz import merge
import numpy as np
import pandas as pd

from marslab.imgops.debayer import make_bayer, debayer_upsample
from marslab.imgops.imgutils import absolutely_destroy, mapfilter, \
    get_all_bands, get_all_bands_from_all
from marslab.imgops.look import Look, save_plainly
from marslab.poolutils import wait_for_it

log = logging.getLogger(__name__)


class BandSet:
    """
    class for organizing and performing bulk rendering operations on
    multispectral image products
    """

    def __init__(
        self,
        metadata: pd.DataFrame = None,
        rois: MutableMapping = None,
        bayer_info: MutableMapping = None,
        load_method: Callable = None,
        name: str = None,
        threads: Mapping = None,
        raw: MutableMapping = None,
        special_constants: Collection[Any] = None
    ):
        """
        :param metadata: dataframe containing at least "PATH", "BAND", "IX,
        " optionally others
        # :param raw: dictionary of NAME:ndarray or pdr.Data corresponding to
        # unprocessed images
        # :param debayered: dictionary of NAME:ndarray or pdr.Data
        # corresponding to debayered versions of raw images
        # :param looks = dictionary of str - Union[ndarray, mpl.figure.Figure,
        # PIL.Image] -- images generated by
        #     other methods
        # :param thumbs = dictionary of str - Union[ndarray, mpl.figure.Figure,
        # PIL.Image] -- thumbnails
        # :param extended -- pd.DataFrame, extended metadata perhaps plus
        # data, variably definable,
        #     but generally unpivoted metadata for both band level and
        #     region/pointing/spectrum level
        # :param compact -- pd.DataFrame, compact metadata perhaps plus data,
        # variably definable,
        #     but generally metadata pivoted on region/pointing/spectrum level
        # :param summary -- pd.DataFrame, just metadata pivoted on
        # pointing/spectrum level
        :param rois -- str, pathlike, or hdulist with 'regions of interest'
        drawn on images
        :param bayer_info: dict containing one or more of 'mask': full bayer
        masks fitting raw image sizes,
            'pattern' -- bayer pattern definition, 'row_column' -- row and
            column positions for
            bilinear_interpolate_subgrid
        :param load_method -- 'pdr' or 'rasterio' -- how to open files
        :param name -- str, designated name for this observation / cluster
        of bands / etc.
        # :param counts -- df of values for ROIs counted across bands
        :param threads -- dict of thread counts for different things
        """
        self.metadata = metadata
        self.raw = raw
        self.debayered = None
        self.looks = None
        self.thumbs = None
        self.extended = None
        self.compact = None
        self.summary = None
        self.rois = rois
        self.bayer_info = bayer_info
        self.load_method = load_method
        self.name = name
        self.counts = None
        self.threads = threads if threads is not None else {}
        # optional cache of auxiliary objects. loaders can use this when
        # complex metadata have been pre-parsed, etc..
        self.precached = {}
        self.local_files = []
        self.special_constants = special_constants
        if isinstance(metadata, pd.DataFrame):
            if "IX" not in metadata.columns:
                metadata["IX"] = 0
        for mapping in ("raw", "looks", "thumbs", "debayered", "threads"):
            if getattr(self, mapping) is None:
                setattr(self, mapping, {})
        self.cache_names = ("raw", "debayered", "looks")
        if self.load_method is None:
            from marslab.imgops.loaders import pil_load

            self.load_method = pil_load

    # TODO: do I need to allow more of these on init? like for copying? maybe?

    def setup_pool(self, thread_type):
        if self.threads.get(thread_type) is None:
            return None
        from pathos.multiprocessing import ProcessPool

        log.info("... initializing worker pool ...")
        pool = ProcessPool(self.threads.get(thread_type))
        pool.restart()
        return pool

    def load(
        self, bands: Collection[str], reload: bool = False, quiet: bool = False
    ):
        pool = self.setup_pool("load")
        load_df = self.metadata.copy()
        if bands == "all":
            bands = load_df["BAND"]
        else:
            bands = pd.Series(tuple(bands))
        if reload is False:
            bands = bands.loc[~bands.isin(self.raw.keys())]
        if (quiet is False) and not (bands.isin(load_df["BAND"]).all()):
            log.info("Not all requested bands are available.")
        # group bands by file -- for instruments like MASTCAM, this is a
        # one-row / band df per file; for RGB images, three rows per file; for
        # instruments like Kaguya or Supercam, dozens or hundreds per file.
        loading = load_df.loc[load_df["BAND"].isin(bands)]
        chunked_by_file = loading.dropna(subset=["PATH"]).groupby("PATH")
        # TODO, maybe: dispatch single and multithreaded cases separately?
        if pool is None:
            results = []
            for path, band_df in chunked_by_file:
                results.append(
                    self.load_method(path, band_df, bands, self.precached))
                log.info("loaded " + path)
        else:
            results = {}
            # caution: dict comprehension does _not_ work well here
            for path, band_df in chunked_by_file:
                results[path] = pool.apipe(
                    self.load_method, path, band_df, bands
                )
            results = wait_for_it(pool, results, log)
        self.raw |= merge(results)

    def make_db_masks(self, shape: Sequence[int, int] = None, remake=False):
        if self.bayer_info is None:
            raise ValueError("can't make debayer masks with no debayer info.")
        if "masks" in self.bayer_info.keys():
            if remake is False:
                return
        if shape is None:
            try:
                shape = next(iter(self.raw.values())).shape
            except (AttributeError, StopIteration):
                raise ValueError(
                    "Need loaded images or an explicit shape to make debayer "
                    "masks."
                )
        self.bayer_info["masks"] = make_bayer(
            shape, self.bayer_info["pattern"]
        )
        self.bayer_info["row_column"] = {
            pixel: (np.unique(mask[0]), np.unique(mask[1]))
            for pixel, mask in self.bayer_info["masks"].items()
        }

    def bayer_pixel(self, band_name: str):
        try:
            return self.metadata.loc[
                self.metadata["BAND"] == band_name, "BAYER_PIXEL"
            ].iloc[0]
        except (KeyError, ValueError, AttributeError, IndexError):
            return None

    def wavelength(self, band_names: Union[str, Sequence[str]]):
        wavelengths = []
        if isinstance(band_names, str):
            band_names = [band_names]
        for band_name in band_names:
            try:
                wavelengths.append(
                    self.metadata.loc[
                        self.metadata["BAND"] == band_name, "WAVELENGTH"
                    ].iloc[0]
                )
            except (KeyError, ValueError, AttributeError, IndexError):
                continue
        return wavelengths

    def debayer_if_required(self, band_name: str, use_cache: bool = True):
        """
        return a debayered version of an image, if self.metadata["BAYER_PIXEL"]
        suggests that there's debayering to do and a raw image is available.
        optionally fetch from cache of debayered images.
        """
        pixel = self.bayer_pixel(band_name)
        if pixel is None:
            return None
        if use_cache and (band_name in self.debayered.keys()):
            return self.debayered[band_name]
        self.make_db_masks()
        return debayer_upsample(
            self.raw[band_name],
            pixel=pixel,
            masks=self.bayer_info["masks"],
            row_column=self.bayer_info["row_column"],
        )

    def bulk_debayer(self, bands: Collection[str]):
        """
        debayer all bands according to spec in self.metadata and
        self.bayer_info, asynchronously / multithreaded if
        bandset.threads["debayer"] is set; cache in self.debayered

        don't set None for non-debayered images: debayer availability
        should be visible by looking at bandset.debayered's keys
        """
        if bands == "all":
            bands = self.metadata["BAND"].unique()
        for band in bands:
            debayer = self.debayer_if_required(band)
            if debayer is not None:
                self.debayered[band] = debayer
        # TODO: threading isn't working. this is not presently a performance
        #  concern because debayering is not a serious bottleneck for things
        #  we're doing
        # pool = self.setup_pool("debayer")
        # if pool is not None:
        #     db = pool.map_async(self.debayer_if_required, bands)
        #     pool.close()
        #     pool.join()
        #     # this is like the unthreaded case
        #     # except that we get the debayer object as a chunk
        #     self.debayered = {
        #         band: array_or_none
        #         for band, array_or_none in zip(
        #             filter(lambda x: x is not None, bands), db.get()
        #         )
        #     }
        #     return

    def get_band(self, band: str):
        """
        get the "most processed" (right now just meaning debayered if
        in a bayered band, raw if not in a bayered band) version of a cached
        image. does not 'intelligently' check for anything like
        BandSet.debayer_if_required() -- this grabs _only_ from cache.
        caveat emptor. TODO: consider unifying these anyway.
        """
        if band in self.debayered.keys():
            return self.debayered[band]
        return self.raw[band]

    def prep_look_set(self, instructions: Collection[Mapping], autoload: bool):
        """
        filter the instruction set we want for the images we have.
        if requested, also load/cache images, debayering as required.
        """
        # what bands do we want?
        desired_bands = get_all_bands_from_all(instructions)
        # try to get them, if we don't have them yet
        if (autoload is True) and (self.metadata is not None):
            self.load(desired_bands, quiet=True)
            # debayer them, if needed
            self.bulk_debayer(
                set(self.raw.keys()).intersection(desired_bands),
            )
        # what looks can we make with what we have?
        available_looks = mapfilter(
            lambda instruction: set(get_all_bands(instruction)).issubset(
                tuple(self.raw.keys())
            ),
            None,
            instructions,
        )
        for op in [op for op in instructions if op not in available_looks]:
            log.info(
                "skipping " + str(op.get("name")) + " due to missing bands"
            )
        return available_looks

    def make_look_set(
        self,
        instructions: Collection[Mapping],
        autoload: bool = True,
        timeout: Optional[float] = None
    ):
        # load images and filter instruction set for unavailable bands
        available_instructions = self.prep_look_set(instructions, autoload)
        # TODO, maybe: print skipping messages
        look_cache = {}
        pool = self.setup_pool("look")
        if pool is not None:
            log.info("... serializing arrays ...")
        for instruction in available_instructions:
            # do we have a special name? TODO: make this more opinionated?
            op_name = instruction.get("name")
            if op_name is None:
                op_name = (
                    instruction["look"] + "_" + "_".join(instruction["bands"])
                )
            if instruction["look"] not in ["nested_composite"]:
                op_images = [
                    self.get_band(band).copy() for band in instruction["bands"]
                ]
            else:
                # TODO: maybe generalize...and/or dispatch this all better.
                op_images = {
                    channel: [
                        self.get_band(band).copy()
                        for band in instruction['params'][channel]["bands"]
                    ]
                    for channel in ("red", "green", "blue")
                }
            special_kwargs = {}
            # make processing pipeline from instruction
            # note that wavelengths for spectops are added automagically by
            # the Look compiler.
            pipeline = Look.compile_from_instruction(
                instruction,
                metadata=self.metadata,
                special_constants=self.special_constants
            )
            if "underlay" in instruction.keys():
                underlay = instruction['underlay']['band']
                if isinstance(underlay, int):
                    pipeline.add_underlay(op_images[underlay])
                else:
                    pipeline.add_underlay(self.get_band(underlay))
            if pool is not None:
                look_cache[op_name] = pool.apipe(
                    pipeline.execute, op_images, **special_kwargs
                )
            else:
                look_cache[op_name] = pipeline.execute(
                    op_images, **special_kwargs
                )
                log.info("generated " + op_name)
        if pool is not None:
            look_cache = wait_for_it(
                pool,
                look_cache,
                log,
                message="generated ",
                as_dict=True,
                timeout=timeout
            )
            pool.terminate()
        self.looks |= look_cache

    def purge(self, what: Optional[str] = None) -> None:
        if what == "precached_images":
            for data in self.precached.values():
                del data.IMAGE
            return
        if what is None:
            for cache_name in self.cache_names:
                absolutely_destroy(getattr(self, cache_name), True)
                setattr(self, cache_name, {})
        elif what in self.cache_names:
            absolutely_destroy(getattr(self, what), True)
            setattr(self, what, {})
        else:
            raise ValueError(str(what) + " is not a valid cache type.")
        gc.collect()

    @staticmethod
    def write_plain_image(look, look_name, outpath, pool, prefix, results):
        filename = look_name + ".png"
        if prefix is not None:
            filename = f"{prefix} {filename}"
        if pool is None:
            save_plainly(look, filename, outpath)
            log.info("wrote " + filename)
        else:
            results[filename] = pool.apipe(
                save_plainly, look, filename, outpath
            )
        return filename

    def save_looks(self, outpath, prefix=None, threads=None):
        if prefix is None:
            prefix = self.name
        pool = None
        results = {}
        # TODO: dispatch these cases
        if threads is not None:
            from pathos.multiprocessing import ProcessPool

            log.info("... initializing worker pool ...")
            pool = ProcessPool(threads)
            pool.restart()
            log.info("... serializing images ...")
        for look_name, look in self.looks.items():
            # TODO: ugh.
            image_path = str(outpath)
            if not os.path.exists(image_path):
                os.makedirs(image_path)
            filename = self.write_plain_image(
                look, look_name, image_path, pool, prefix, results
            )
            self.local_files.append(str(Path(image_path, filename)))
        if pool is not None:
            # TODO: extend this, generally speaking, to give useful messages
            #  about failure
            wait_for_it(pool, results, log, "wrote ")
            pool.terminate()


class ImageBands(BandSet):
    """
    simple case of a bandset produced from a single multichannel image.
    """

    def __init__(self, path, load_method=None, **bandset_kwargs):
        if load_method is None:
            from marslab.imgops.loaders import pil_load

            load_method = pil_load
        metadata = pd.DataFrame()
        super().__init__(
            metadata=metadata, load_method=load_method, **bandset_kwargs
        )
        # TODO: stuff about automatically coercing grayscale if desired
        self.raw = load_method(path)
        metadata["BAND"] = self.raw.keys()
        metadata["IX"] = self.raw.keys()
        metadata["PATH"] = path





"""
fast, numerical, non-vectorized version of slowtime.py 
eschewing external dependencies other than dateutil.parser

TODO: move / copy references from that module here
"""

import curses
import datetime as dt
from itertools import starmap
from math import cos, floor, sin, pi
from typing import Callable, Union

import dateutil.parser as dtp

TZS = {
    'EST': -18000,
    'CST': -21600,
    'MST': -25200,
    'PST': -28800
}

LEAP_SECOND_THRESHOLDS = [
    dt.datetime(1972, 6, 30, tzinfo=dt.timezone.utc),
    dt.datetime(1972, 12, 31, tzinfo=dt.timezone.utc),
    dt.datetime(1973, 12, 31, tzinfo=dt.timezone.utc),
    dt.datetime(1974, 12, 31, tzinfo=dt.timezone.utc),
    dt.datetime(1975, 12, 31, tzinfo=dt.timezone.utc),
    dt.datetime(1976, 12, 31, tzinfo=dt.timezone.utc),
    dt.datetime(1977, 12, 31, tzinfo=dt.timezone.utc),
    dt.datetime(1978, 12, 31, tzinfo=dt.timezone.utc),
    dt.datetime(1979, 12, 31, tzinfo=dt.timezone.utc),
    dt.datetime(1981, 6, 30, tzinfo=dt.timezone.utc),
    dt.datetime(1982, 6, 30, tzinfo=dt.timezone.utc),
    dt.datetime(1983, 6, 30, tzinfo=dt.timezone.utc),
    dt.datetime(1985, 6, 30, tzinfo=dt.timezone.utc),
    dt.datetime(1987, 12, 31, tzinfo=dt.timezone.utc),
    dt.datetime(1989, 12, 31, tzinfo=dt.timezone.utc),
    dt.datetime(1990, 12, 31, tzinfo=dt.timezone.utc),
    dt.datetime(1992, 6, 30, tzinfo=dt.timezone.utc),
    dt.datetime(1993, 6, 30, tzinfo=dt.timezone.utc),
    dt.datetime(1994, 6, 30, tzinfo=dt.timezone.utc),
    dt.datetime(1995, 12, 31, tzinfo=dt.timezone.utc),
    dt.datetime(1997, 6, 30, tzinfo=dt.timezone.utc),
    dt.datetime(1998, 12, 31, tzinfo=dt.timezone.utc),
    dt.datetime(2005, 12, 31, tzinfo=dt.timezone.utc),
    dt.datetime(2008, 12, 31, tzinfo=dt.timezone.utc),
    dt.datetime(2012, 6, 30, tzinfo=dt.timezone.utc),
    dt.datetime(2015, 6, 30, tzinfo=dt.timezone.utc),
    dt.datetime(2016, 12, 31, tzinfo=dt.timezone.utc)
]


# algorithm derived from Julian Date article on scienceworld.wolfram.com
# itself apparently based on Danby, J. M., Fundamentals of Celestial Mechanics
def dt_to_jd(time: dt.datetime) -> float:
    y, m, d = time.year, time.month, time.day
    h = time.hour + time.minute / 60 + time.second / 3600
    return sum([
        367 * y,
        -1 * floor(7 * (y + floor((m + 9) / 12)) / 4),
        -1 * floor(3 * (floor((y + (m - 9) / 7) / 100) + 1) / 4),
        floor(275 * m / 9) + d + 1721028.5,
        h / 24
    ])


def utc_to_tt_offset(time: dt.datetime) -> float:
    """
    return number of seconds necessary to advance UTC
    to TT. we aren't presently supporting dates prior to 1972
    because fractional leap second handling is another thing.
    """
    if time.year < 1972:
        raise ValueError("dates prior to 1972 are not currently supported")
    # this includes the horrible fractional leap seconds prior to 1972
    # and the base 32.184 s offset between TAI and TT
    offset = 42.184
    for threshold in LEAP_SECOND_THRESHOLDS:
        if time > threshold:
            offset += 1
    return offset


def to_jd_tt(time: Union[dt.datetime, str, None] = None) -> float:
    """
    transforms a UTC python datetime -- or any string representing UTC
    time in a format dateutil can parse -- into a julian day number
    in terrestrial time
    """
    if time is None:
        time = dt.datetime.utcnow()
    elif isinstance(time, str):
        time = dtp.parse(time, tzinfos=TZS)
    if time.tzinfo is None:
        time = time.replace(tzinfo=dt.timezone.utc)
    time = time.astimezone(dt.timezone.utc)
    return dt_to_jd(time) + utc_to_tt_offset(time) / 86400


def to_j2000_jd(jd: float) -> float:
    """
    time difference between time and J2000 epoch in julian days.
    """
    return jd - 2451545


def r2d(radian_value: float) -> float:
    """radians to degrees"""
    return radian_value * 180 / pi


def hours_to_24h_time(hours: float) -> str:
    """
    convoluted expression because of Python's...thing to transform a figure
    given in fractional hours to a string in ISO time format
    """
    return (
            dt.datetime(2001, 1, 2) + dt.timedelta(hours=hours)
    ).time().isoformat()


def mean_anomaly(dt_j2000: float) -> float:
    """eq. 16 a&m"""
    return 0.3383687274133927 + 0.009145887087204227 * dt_j2000


def fictitious_mean_sun(dt_j2000: float) -> float:
    """eq. 17 a&m"""
    return 4.719145149919159 + 0.009146197162399134 * dt_j2000


# table 5 (truncated)
PERTURBATION_TABLE = [
    [1.23918377e-04, 2.23530000e+00, 8.62349730e-01],
    [9.94837674e-05, 2.75430000e+00, 2.93517256e+00],
    [6.80678408e-05, 1.11770000e+00, 3.34818728e+00],
    [6.45771823e-05, 1.57866000e+01, 3.79364766e-01],
    [3.66519143e-05, 2.13540000e+00, 2.74086506e-01],
    [3.49065850e-05, 2.46940000e+00, 1.66727813e+00],
    [3.14159265e-05, 3.28493000e+01, 8.56869396e-01]
]


def perturbation_factory(dt_j2000: float) -> Callable[[float, float, float], float]:
    """
    generate the interior of the perturbation summation
    """

    def calculate_perturbation(a: float, tau: float, phi: float) -> float:
        return a * cos(
            0.01720241889326163 * dt_j2000 / tau + phi
        )

    return calculate_perturbation


def pbs(dt_j2000: float) -> float:
    """calculate the sum of perturbations"""
    return sum(starmap(perturbation_factory(dt_j2000), PERTURBATION_TABLE))


def mean_terms(dt_j2000: float) -> float:
    """terms of eq. 19"""
    # writing it in this silly way for readability
    alpha_m = mean_anomaly(dt_j2000)
    return sum([
        (0.1865931503307138 + 5.235987755982988e-09 * dt_j2000) * sin(alpha_m),
        0.010873401239924672 * sin(2 * alpha_m),
        0.0008726646259971648 * sin(3 * alpha_m),
        8.726646259971648e-05 * sin(4 * alpha_m),
        8.726646259971648e-06 * sin(5 * alpha_m)
    ])


# areocentric solar longitude
def l_s(dt_j2000: float) -> float:
    return fictitious_mean_sun(dt_j2000) + mean_terms(dt_j2000) + pbs(dt_j2000)


def eot(dt_j2000: float) -> float:
    l_s_0 = l_s(dt_j2000)
    return sum([
        0.04993386989955777 * sin(2 * l_s_0),
        -0.001239183768915974 * sin(4 * l_s_0),
        3.490658503988659e-05 * sin(6 * l_s_0),
        fictitious_mean_sun(dt_j2000),
        -1 * l_s_0
    ])


def mst_hours(julian_day: float) -> float:
    days = (julian_day - 2451549.5) / 1.0274912517 + 44796.0 - 0.0009626
    return (days * 24) % 24


def lmst_hours(julian_day: float, west_longitude_degrees: float) -> float:
    return mst_hours(julian_day) - west_longitude_degrees * 24 / 360


def ltst_hours(julian_day: float, west_longitude_degrees: float) -> float:
    hours = lmst_hours(julian_day, west_longitude_degrees) + \
            r2d(eot((to_j2000_jd(julian_day)))) / 15
    return hours


def mst(time: Union[str, None, dt.datetime] = None) -> str:
    jd_tt = to_jd_tt(time)
    return hours_to_24h_time(
        mst_hours(
            jd_tt
        )
    )


def lmst(
        time: Union[str, None, dt.datetime] = None,
        west_longitude: float = 0
) -> str:
    jd_tt = to_jd_tt(time)
    return hours_to_24h_time(
        lmst_hours(
            jd_tt,
            west_longitude
        )
    )


def ltst(
        time: Union[str, None, dt.datetime] = None,
        west_longitude: float = 0
) -> str:
    jd_tt = to_jd_tt(time)
    return hours_to_24h_time(
        ltst_hours(
            jd_tt,
            west_longitude
        )
    )


mission_to_wlon = {
    "Phoenix": 360 - 234.250778,
    "MSL": 360 - 137.4417,
    "InSight": 360 - 135.623,
    "M2020": 360 - 77.5945,
    "AMT": 0,
}

time_type_to_function = {
    'ltst': ltst,
    'lmst': lmst
}


def drawclock(stdscr, mission: str = 'M2020', time_type='ltst'):
    k = 0
    curses.curs_set(0)
    stdscr.nodelay(True)
    stdscr.clear()
    stdscr.refresh()
    stdscr.nodelay(True)
    time_function = time_type_to_function[time_type]
    while (k != ord('q')):
        stdscr.addstr(
            1, 1,
            f'{mission} - {time_function(west_longitude=mission_to_wlon[mission])[:8]}'
        )
        stdscr.refresh()
        curses.delay_output(100)
        k = stdscr.getch()

    curses.nocbreak()
    stdscr.keypad(False)
    curses.echo()




from scipy.ndimage import gaussian_filter

from marslab.imgops.imgutils import std_clip
from marslab.imgops.render import colormapped_plot


def gen_spectop_defaults(special_constants=None):
    return {
        "params": {"special_constants": special_constants},
        "limiter": {"function": std_clip},
        "postfilter": {"function": gaussian_filter, "params": {"sigma": 2}},
        "plotter": {
            "function": colormapped_plot,
            "params": {
                "cmap": "orange_teal",
                "render_colorbar": True,
                "special_constants": special_constants,
            },
        },
    }



"""
utilities for composing lightweight imaging pipelines
"""
from abc import ABC
from collections.abc import Callable, Mapping, Sequence
from functools import partial
from inspect import signature
from pathlib import Path
from typing import TYPE_CHECKING, Optional, Collection, Any

from dustgoggles.composition import Composition
from dustgoggles.func import argstop

import marslab.spectops
from marslab.imgops.imgutils import map_filter, crop_all

if TYPE_CHECKING:
    import numpy as np
    import pandas as pd


def look_to_function(look: str) -> Callable:
    from marslab.imgops import render
    
    if look in marslab.spectops.SPECTOP_NAMES:
        return partial(
            render.spectop_look, spectop=getattr(marslab.spectops, look)
        )
    elif look in ("enhanced color", "true color", "composite"):
        return render.render_rgb_composite
    elif look == "dcs":
        return render.decorrelation_stretch
    elif look == "nested_composite":
        return render.render_nested_rgb_composite
    else:
        raise ValueError("unknown look operation " + look)


def interpret_look_step(instruction):
    """
    heart of look pipeline: the look function (spectrum op, dcs, etc.)
    note this step is _mandatory_.
    """
    try:
        step = instruction["look"]
        if isinstance(step, str):
            step = look_to_function(step)
    except KeyError:
        raise ValueError("The instruction must include at least a look type.")
    return step, instruction.get("params", {})


def interpret_crop_step(bounds):
    """
    note at present a cropper is always included whether you ask for it or not,
    but by default it acts as the identity function.
    """
    return crop_all, {"bounds": bounds}


def interpret_prefilter_step(chunk):
    step = chunk["function"]
    # by default, wrap the prefilter function so that it works as a variadic
    # function, applying itself to each input channel individually; some
    # notional prefilters may of course not want this!
    if chunk.get("map") is not False:
        step = map_filter(step)
    return step, chunk.get("params", {})


def interpret_mask_step(chunk):
    from marslab.imgops.masking import extract_masks
    return extract_masks, chunk


def interpret_instruction_step(
    instruction: Mapping, step_name: str
) -> tuple[Optional[Callable], Mapping]:
    """
    unpack individual elements of the look instruction markup
    syntax into steps and kwargs for a Composition.
    """
    # check central mandatory element first
    if step_name == "look":
        return interpret_look_step(instruction)
    # bail out if non-mandatory element is not present
    if step_name not in instruction.keys():
        return None, {}
    chunk = instruction[step_name]
    # other special cases
    if step_name == "crop":
        return interpret_crop_step(chunk)
    if step_name == "prefilter":
        return interpret_prefilter_step(chunk)
    if step_name == "mask":
        return interpret_mask_step(chunk)
    # specifying a function is mandatory,
    # specifying bound parameters is not
    return chunk["function"], chunk.get("params", {}).copy()


class Look(Composition, ABC):
    def __init__(
        self,
        *args,
        metadata: "pd.DataFrame" = None,
        bands: tuple[str] = None,
        special_constants: Collection[Any] = None,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self.metadata = metadata
        self.bands = bands
        self.special_constants = special_constants
        if (self.metadata is not None) or (self.special_constants is not None):
            self.populate_kwargs()

    def _add_wavelengths(self, wavelengths: Sequence[float]):
        look = self.steps["look"]
        if "__name__" in dir(look):
            name = look.__name__
        else:
            name = look.func.__name__
        if name != "spectop_look":
            return False
        self.add_insert("look", "wavelengths", wavelengths)
        return True

    def _get_plotter_layers(self):
        try:
            return self.inserts['plotter']['layers']
        except KeyError:
            self.add_insert('plotter', 'layers', [])
            return self._get_plotter_layers()

    def add_underlay(self, underlay: "np.ndarray", layer_ix: int = -1):
        if "crop" in self.steps:
            underlay = self.steps["crop"](
                underlay, **self.inserts.get('crop', {})
            )
        self._get_plotter_layers().append(
            {'layer_ix': layer_ix, "image": underlay}
        )

    @classmethod
    def compile_from_instruction(
        cls,
        instruction: Mapping,
        metadata: "pd.DataFrame" = None,
        special_constants: Collection[Any] = None
    ):
        """compile a look instruction into a rendering pipeline"""
        # only look is required. these are _possible_ steps in the pipeline.
        step_names = (
            "crop",
            "prefilter",
            "mask",
            "look",
            "limiter",
            "postfilter",
            "plotter",
            "bang",
        )
        steps = {}
        inserts = {}
        for step_name in step_names:
            step, kwargs = interpret_instruction_step(instruction, step_name)
            if step is not None:
                steps[step_name] = step
            if kwargs != {}:
                inserts[step_name] = kwargs
        look = cls(
            steps,
            inserts=inserts,
            metadata=metadata,
            special_constants=special_constants,
            bands=instruction.get("bands"),
        )
        if "mask" in steps.keys():
            # add 'splitter'
            # TODO -- maybe implement this at Composition level
            look.steps['look'] = argstop(look.steps['look'], 2)
            look.add_send('mask', lambda s: s[2], look._get_plotter_layers())
        return look

    # TODO: is this excessively baroque; would an internal dispatch be better?
    def populate_kwargs(self):
        for step in self.steps:
            params = signature(self.steps[step]).parameters.values()
            param_names = [param.name for param in params]
            for thing in ("special_constants", "metadata"):
                if (hasattr(self, thing)) and (thing in param_names):
                    self.add_insert(step, thing, getattr(self, thing))
        if self.bands is not None:
            if "WAVELENGTH" in self.metadata.columns:
                wavelengths = []
                for band in self.bands:
                    try:
                        wavelengths.append(
                            self.metadata.loc[
                                self.metadata["BAND"] == band, "WAVELENGTH"
                            ].iloc[0]
                        )
                    except (KeyError, IndexError):
                        continue
                self._add_wavelengths(wavelengths)


def save_plainly(look, filename, outpath, dpi=275):
    from matplotlib.figure import Figure

    if isinstance(look, Figure):
        for ix, axis in enumerate(look.axes):
            if ix > 0:
                axis.remove()
            else:
                axis.axis("off")
        look.savefig(
            Path(outpath, filename), dpi=dpi, bbox_inches="tight", pad_inches=0
        )
    else:
        look.save(Path(outpath, filename))



"""
Constants and utility functions for compatibility with MERTools.

MERTools is an IDL software suite used for tactical data assessment and derived
product generation. It is the exclusive intellectual property of Arizona State
University (PI J. Bell) and made available on an as-needed basis.
"""

from itertools import chain

import numpy as np
import pandas as pd

from marslab.compat.xcam import (
    make_xcam_filter_dict,
    WAVELENGTH_TO_FILTER,
)

# mappings between MERspect color designations and hex codes
MERSPECT_M20_COLOR_MAPPINGS = {
    "green": "#00ff3b",  # not a creative color
    "yellow": "#eeff00",
    "blue": "#1500ff",
    "red": "#ff0000",
    "magenta": "#ff00ae",
    "cyan": "#00ffee",
    "orange": "#ffa100",
    "azure": "#0094ff",
    "purple": "#ad0bbf",
    "lime": "#9bec11",
    "rust": "#dd5b10",
    "green+2": "#7fff9d",
    "green-1": "#00b229",
    "green-2": "#007f1d",
    "yellow-2": "#777f00",
    "blue+2": "#8a7fff",
    "blue-1": "#0e00b2",
    "blue-2": "#0a007f",
    "red+2": "#ff7f7f",
    "red-1": "#b20000",
    "red-2": "#7f0000",
    "magenta+2": "#ff7fd6",
    "magenta+1": "#ff4cc6",
    "magenta-1": "#b20079",
    "magenta-2": "#7f0057",
    "magenta-3": "#4c0034",
    "cyan+2": "#a6fff9",
    "cyan+1": "#73fff6",
    "cyan-1": "#00b2a6",
    "cyan-2": "#007f77",
    "cyan-3": "#004c47",
    "orange+2": "#ffd07f",
    "orange+1": "#ffbd4c",
    "orange-1": "#b27100",
    "orange-2": "#7f5000",
    "orange-3": "#4c3000",
    "azure+2": "#7fc9ff",
    "azure+1": "#4cb4ff",
    "azure-1": "#0068b2",
    "azure-2": "#004a7f",
    "azure-3": "#002c4c",
}
# The colors below here are from the MCAM / PCAM legacy functionality

MERSPECT_MSL_COLOR_MAPPINGS = {
    "red": "#dc143c",  # duplicated from above w/ different hex code
    "light green": "#7fff00",
    "light blue": "#0000ff",
    "light cyan": "#00ffff",
    "dark green": "#128012",
    "yellow": "#ffff00",  # duplicated from above w/ different hex code
    "light purple": "#ff00ff",
    "pink": "#fa8072",
    "teal": "#008080",
    "goldenrod": "#B8860b",
    "sienna": "#a0522d",
    "dark blue": "#000080",
    "bright red": "#ff0000",
    "dark red": "#800000",
    "dark purple": "#800080",
}

MERSPECT_COLOR_MAPPINGS = {
    **MERSPECT_M20_COLOR_MAPPINGS,
    **MERSPECT_MSL_COLOR_MAPPINGS,
}

# Standard assignments of MERSpect ROI colors to feature classes.
# for the Western Washington University Reflectance Lab (PI M. Rice).
# No such mapping has yet been agreed upon for ZCAM.
COLOR_TO_FEATURE_TYPE = {
    "MCAM": {
        "light purple": "drill tailings",
        "dark purple": "dump piles",
        "light blue": "dusty rock",
        "teal": "dusty rock",
        "dark blue": "dusty rock",
        "light green": "DRT target",
        "dark green": "broken rock face",
        "bright red": "undisturbed soil",
        "dark red": "disturbed soil",
        "red": "undisturbed soil",
        "light cyan": "nodule-rich rock",
        "goldenrod": "veins",
        "sienna": None,
        "pink": None,
        "yellow": None,
    },
    "ZCAM": {
        "green": None,
        "yellow": None,
        "blue": None,
        "red": None,
        "magenta": None,
        "cyan": None,
        "orange": None,
        "azure": None,
        "purple": None,
        "lime": None,
        "rust": None,
        "green+2": None,
        "green-1": None,
        "green-2": None,
        "yellow-2": None,
        "blue+2": None,
        "blue-1": None,
        "blue-2": None,
        "red+2": None,
        "red-1": None,
        "red-2": None,
        "magenta+2": None,
        "magenta+1": None,
        "magenta-1": None,
        "magenta-2": None,
        "magenta-3": None,
        "cyan+2": None,
        "cyan+1": None,
        "cyan-1": None,
        "cyan-2": None,
        "cyan-3": None,
        "orange+2": None,
        "orange+1": None,
        "orange-1": None,
        "orange-2": None,
        "orange-3": None,
        "azure+2": None,
        "azure+1": None,
        "azure-1": None,
        "azure-2": None,
        "azure-3": None,
    },
}


def parse_merspect_fn(fn):
    # Parse the MERSpect filename for obs information
    sol = int(fn.split("/")[-1].split("_")[0][3:])
    instrument = str.upper(fn.split("/")[-1].split("_")[1][:4])
    seq_id = fn.split("/")[-1].split("_")[1]
    return {"SOL": sol, "INSTRUMENT": instrument, "SEQ_ID": seq_id}


def get_unique_colors(csv, instrument):
    # Figure out the unique color names contained in the MERSpect output file
    colors = []
    for k in csv.keys():
        keycolor = " ".join(k.strip().split(" ")[:-2])
        for c in COLOR_TO_FEATURE_TYPE[instrument].keys():
            if keycolor == c:
                colors += [c]
    return np.unique(colors)


def merspect_to_marslab(fn, write=True):
    csv = pd.read_csv(fn)
    # TODO: find a different way to get the sol, seq_id, instrument from the
    #  file because this isn't actually standardized within MERspect, only
    #  Melissa's lab
    #  -- or, make a clean version that doesn't even try to parse this
    #  because it's never going to be called directly

    # Get the sol, seq_id, instrument from the filename
    try:
        obsparams = parse_merspect_fn(fn)
    # someone has named this something weird -- try to parse the data anyway
    except (ValueError, KeyError):
        obsparams = {"SOL": "-", "INSTRUMENT": "ZCAM", "SEQ_ID": "-"}
    # Rename "# Wavelength (nm)" to "Wavelength (nm)"
    csv.rename(columns={"# Wavelength (nm)": "Wavelength"}, inplace=True)
    # Clean up column names
    [csv.rename(columns={k: k.strip()}, inplace=True) for k in csv.keys()]

    # We want the columns in order of ascending wavelength, regardless of
    # instrument
    columns = {
        "MCAM": [
            "SOL",
            "SEQ_ID",
            "INSTRUMENT",
            "COLOR",
            "FEATURE",
            "FORMATION",
            "MEMBER",
            "FLOAT",
            *list(
                chain.from_iterable(
                    [
                        (filt, filt + "_STD")
                        for filt in make_xcam_filter_dict("MCAM").keys()
                    ]
                )
            ),
        ],
        "ZCAM": [
            "SOL",
            "SEQ_ID",
            "INSTRUMENT",
            "COLOR",
            "FEATURE",
            "FORMATION",
            "MEMBER",
            "FLOAT",
            "RMS",
            "ZOOM",
            *list(
                chain.from_iterable(
                    [
                        (filt, filt + "_STD")
                        for filt in make_xcam_filter_dict("ZCAM").keys()
                    ]
                )
            ),
        ],
    }[obsparams["INSTRUMENT"]]

    # Init the dataframe
    data = pd.DataFrame(columns=columns)
    instrument = obsparams["INSTRUMENT"]

    # Generate the index column --- 'COLOR'
    colors = get_unique_colors(csv, instrument)
    data["COLOR"] = colors
    data = data.set_index("COLOR")

    for color in colors:
        this_color = csv[
            [
                "Eye",
                "Wavelength",
                f"{color} Mean Value",
                f"{color} Standard Deviation",
            ]
        ]
        feature = COLOR_TO_FEATURE_TYPE[instrument].get(color)
        for i in range(len(this_color)):
            try:
                eye = this_color.loc[i]["Eye"].strip()[0]
                wavelength = int(this_color.loc[i]["Wavelength"])
                filt = WAVELENGTH_TO_FILTER[instrument][eye][wavelength]
                data[f"{filt}"].loc[color] = this_color.loc[i][
                    f"{color} Mean Value"
                ]
                data[f"{filt}_STD"].loc[color] = this_color.loc[i][
                    f"{color} Standard Deviation"
                ]
                if feature is not None:
                    data["FEATURE"].loc[color] = feature
            except AttributeError:  # empty entries
                try:
                    if (
                        this_color.iloc[i]["Wavelength"].strip() == "Notes"
                    ) and (feature is None):
                        data["FEATURE"].loc[color] = this_color.iloc[i][
                            f"{color} Mean Value"
                        ]
                    elif this_color.iloc[i]["Wavelength"].strip() == "Float":
                        data["FLOAT"].loc[color] = this_color.iloc[i][
                            f"{color} Mean Value"
                        ]
                    elif (
                        this_color.iloc[i]["Wavelength"].strip() == "Formation"
                    ):
                        data["FORMATION"].loc[color] = this_color.iloc[i][
                            f"{color} Mean Value"
                        ]
                    elif this_color.iloc[i]["Wavelength"].strip() == "Member":
                        data["MEMBER"].loc[color] = this_color.iloc[i][
                            f"{color} Mean Value"
                        ]
                # TODO: what was this for?
                except:
                    pass

    # Set values for constant columns

    data["INSTRUMENT"] = obsparams["INSTRUMENT"]
    data["SOL"] = obsparams["SOL"]
    data["SEQ_ID"] = obsparams["SEQ_ID"]
    # Clean up for consistency
    data["FORMATION"].replace("Murray ", "Murray", inplace=True)
    data["FORMATION"].replace("Kiimberly", "Kimberly", inplace=True)
    data["MEMBER"].replace(
        "Knockfarril Hill ", "Knockfarril Hill", inplace=True
    )
    # Clean up some entries for human-readability
    data["FORMATION"].replace(np.nan, "-", inplace=True)
    data["MEMBER"].replace(np.nan, "-", inplace=True)
    data["FLOAT"].replace(np.nan, "N", inplace=True)
    data["FLOAT"].replace("x", "Y", inplace=True)
    data["FLOAT"].replace("float", "N", inplace=True)
    data["FEATURE"].replace(np.nan, "-", inplace=True)

    data.reset_index(inplace=True)
    data = data[columns]

    fn = fn.replace("-BEFORE", "")  # for test files
    outfile = fn.replace(".csv", "-marslab.csv")
    if write:
        data.to_csv(outfile, index=False)
    return data


def roi_color_ix_to_color_name(color_ix: int, instrument: str = "MCAM") -> str:
    """
    convert MERSpect ROI color index to color name
    """
    # this works because colors are listed in the dictionaries in
    # marslab.compat.mertools in the order
    # they are listed in the MERSpect IDL code. No explicit mapping
    # between MERSpect ROI color indices and actual colors exists.
    if instrument in ["MCAM", "PCAM"]:
        return list(MERSPECT_MSL_COLOR_MAPPINGS.keys())[color_ix]
    elif instrument == "ZCAM":
        # The unit offset to the color indices is because "0" is reserved
        # in MERspect as something like "blank canvas." The "erase"
        # capability has a different color associated with it.
        return list(MERSPECT_M20_COLOR_MAPPINGS.keys())[color_ix - 1]
    else:
        raise ValueError(
            "I don't have ROI color information about "
            + instrument
            + ". I know about PCAM, MCAM, and ZCAM."
        )


def add_merspect_colors_to_edgemaps(edgemap_dict):
    for roi_color_eye, edge in edgemap_dict.items():
        color = roi_color_eye.replace(" right", "").replace(" left", "")
        edgemap_dict[roi_color_eye]["color"] = MERSPECT_COLOR_MAPPINGS[color]
    return edgemap_dict



"""
image processing utility functions
"""
import gc
import sys
from functools import partial, reduce
from itertools import chain
from operator import mul
from typing import (
    Callable,
    Sequence,
    MutableMapping,
    Collection,
    Mapping,
    Union,
    MutableSequence,
)

from dustgoggles.structures import dig_for_values
import numpy as np


def absolutely_destroy(thing, delay_collect=False):
    if isinstance(thing, MutableMapping):
        keys = list(thing.keys())
        for key in keys:
            del thing[key]
    elif isinstance(thing, MutableSequence):
        [thing.pop() for _ in range(len(thing))]
    del thing
    if "matplotlib.pyplot" in sys.modules.keys():
        getattr(sys.modules["matplotlib.pyplot"], "close")("all")
    if delay_collect is False:
        gc.collect()


def eightbit(array, stretch=(0, 0)):
    """return an eight-bit version of an array"""
    return np.round(normalize_range(array, (0, 255), stretch)).astype(np.uint8)


def cropmask(array: np.ndarray, bounds=None, copy=True, **_) -> np.ndarray:
    """
    :param array: array to be cropped
    :param bounds: tuple of (left, right, top, bottom) pixels to crop
    """
    if bounds is None:
        return array
    assert len(bounds) == 4  # test for bad inputs
    pixels = [side if side != 0 else None for side in bounds]
    canvas = np.ones(array.shape)
    for value in (1, 3):
        if isinstance(pixels[value], int):
            if pixels[value] > 0:
                pixels[value] = pixels[value] * -1
    canvas[pixels[2]: pixels[3], pixels[0]: pixels[1]] = 0
    if isinstance(array, np.ma.MaskedArray):
        if copy is True:
            array.mask = np.logical_or(canvas, array.mask)
            return array
        return np.ma.masked_array(
            array.data, np.logical_or(canvas, array.mask)
        )
    return np.ma.masked_array(array, canvas)


def crop(array: np.ndarray, bounds=None, **_) -> np.ndarray:
    """
    :param array: array to be cropped
    :param bounds: tuple of (left, right, top, bottom) pixels to crop
    """
    if bounds is None:
        return array
    assert len(bounds) == 4  # test for bad inputs
    pixels = [side if side != 0 else None for side in bounds]
    for value in (1, 3):
        if isinstance(pixels[value], int):
            if pixels[value] > 0:
                pixels[value] = pixels[value] * -1
    return array[pixels[2] : pixels[3], pixels[0] : pixels[1]]


def crop_all(
    arrays: Collection[np.ndarray], bounds=None, **_
) -> Union[Mapping[str, list[np.ndarray]], list[np.ndarray], np.ndarray]:
    """applies crop() to every array in the passed collection"""
    # if you _didn't_ pass it a collection, just overload / dispatch to crop()
    if isinstance(arrays, np.ndarray):
        return crop(arrays, bounds)
    # otherwise map crop()
    elif isinstance(arrays, Mapping):
        return {
            name: [crop(array, bounds) for array in name_arrays]
            for name, name_arrays in arrays.items()
        }
    return [crop(array, bounds) for array in arrays]


def clip_unmasked(array: np.ma.MaskedArray):
    return np.clip(array.data, array.min(), array.max())


def split_filter(filter_function: Callable, axis: int = -1) -> Callable:
    """
    produce a 'split' version of a filter that applies itself to slices across
    a particular axis -- e.g., take a gaussian blur function, return a function
    that applies a gaussian blur to R / G / B channels separately and then
    recomposes them
    """

    def multi(array, *, set_axis: int = axis, **kwargs) -> np.ndarray:
        filt = partial(filter_function, **kwargs)
        filtered = tuple(
            map(filt, np.split(array, array.shape[set_axis], set_axis))
        )
        if isinstance(filtered[0], np.ma.MaskedArray):
            return np.ma.concatenate(filtered, axis=set_axis)
        return np.concatenate(filtered, axis=set_axis)

    return multi


def map_filter(filter_function: Callable) -> Callable:
    """
    returns a version of a function that automatically maps itself across all
    elements of a collection
    """

    def mapped_filter(arrays, *args, **kwargs):
        return [filter_function(array, *args, **kwargs) for array in arrays]

    return mapped_filter


def find_masked_bounds(image, cheat_low, cheat_high):
    """
    relatively memory-efficient way to perform bound calculations for
    normalize_range on a masked array.
    """
    if isinstance(image.mask, np.bool_):
        if image.mask:
            valid = np.array([])
        else:
            valid = image.data
    else:
        valid = image[~image.mask].data
    valid = valid[np.isfinite(valid)]
    if valid.size == 0:
        return None, None
    if (cheat_low != 0) and (cheat_high != 0):
        minimum, maximum = np.percentile(
            valid, [cheat_low, 100 - cheat_high], overwrite_input=True
        ).astype(image.dtype)
    elif cheat_low != 0:
        maximum = valid.max()
        minimum = np.percentile(valid, cheat_low, overwrite_input=True).astype(
            image.dtype
        )
    elif cheat_high != 0:
        minimum = valid.min()
        maximum = np.percentile(
            valid, 100 - cheat_high, overwrite_input=True
        ).astype(image.dtype)
    else:
        minimum = valid.min()
        maximum = valid.max()
    return minimum, maximum


# noinspection PyArgumentList
def find_unmasked_bounds(image, cheat_low, cheat_high):
    """straightforward way to find unmasked array bounds for normalize_range"""
    if cheat_low != 0:
        minimum = np.percentile(image, cheat_low).astype(image.dtype)
    else:
        minimum = image.min()
    if cheat_high != 0:
        maximum = np.percentile(image, 100 - cheat_high).astype(image.dtype)
    else:
        maximum = image.max()
    return minimum, maximum


# TODO: address rollover issues
def normalize_range(
    image: np.ndarray,
    bounds: Sequence[int] = (0, 1),
    stretch: Union[float, tuple[float, float]] = 0,
    inplace: bool = False,
) -> np.ndarray:
    """
    simple linear min-max scaler that optionally percentile-clips the input at
    stretch = (low_percentile, 100 - high_percentile). if inplace is True,
    may transform the original array, with attendant memory savings and
    destructive effects.
    """
    if isinstance(stretch, Sequence):
        cheat_low, cheat_high = stretch
    else:
        cheat_low, cheat_high = (stretch, stretch)
    range_min, range_max = bounds
    if isinstance(image, np.ma.MaskedArray):
        minimum, maximum = find_masked_bounds(image, cheat_low, cheat_high)
        if minimum is None:
            return image
    else:
        minimum, maximum = find_unmasked_bounds(image, cheat_low, cheat_high)
    if not ((cheat_high is None) and (cheat_low is None)):
        if inplace is True:
            image = np.clip(image, minimum, maximum, out=image)
        else:
            image = np.clip(image, minimum, maximum)
    if inplace is True:
        # perform the operation in-place
        image -= minimum
        image *= range_max - range_min
        if image.dtype.char in np.typecodes["AllInteger"]:
            # this loss of precision is probably better than
            # automatically typecasting it.
            # TODO: detect rollover cases, etc.
            image //= maximum - minimum
        else:
            image /= maximum - minimum
        image += range_min
        return image
    return (image - minimum) * (range_max - range_min) / (
        maximum - minimum
    ) + range_min


def enhance_color(image: np.ndarray, bounds, stretch):
    """
    wrapper for normalize_range -- normalize each channel individually,
    conventional "enhanced color" operation
    """
    return split_filter(normalize_range)(image, bounds=bounds, stretch=stretch)


def clip_finite(image, a_min: int, a_max: int):
    finite = np.ma.masked_invalid(image)
    result = np.ma.clip(finite, a_min, a_max)
    if isinstance(image, np.ma.MaskedArray):
        return result
    return result.data


def std_clip(image, sigma=1):
    """
    simple clipping function that clips at multiples of an array's standard
    deviation offset from its mean
    """
    finite = np.ma.masked_invalid(image)
    mean = np.ma.mean(finite)
    std = np.ma.std(finite)
    result = np.ma.clip(finite, *(mean - std * sigma, mean + std * sigma))
    if isinstance(image, np.ma.MaskedArray):
        return result
    return result.data


def centile_clip(image, centiles=(1, 99)):
    """
    simple clipping function that clips values above and below a given
    percentile range
    """
    finite = np.ma.masked_invalid(image)
    bounds = np.percentile(finite[~finite.mask].data, centiles)
    result = np.ma.clip(finite, *bounds)
    if isinstance(image, np.ma.MaskedArray):
        return result
    return result.data


def minmax_clip(image, stretch=(0, 0)):
    """
    simple minmax clip that optionally cheats 0 up and 1 down at multiples
    of an array's dynamic range
    """
    finite = np.ma.masked_invalid(image)
    if stretch != (0, 0):
        dynamic_range = finite.max() - finite.min()
        cheat_low, cheat_high = stretch
    else:
        dynamic_range, cheat_low, cheat_high = (0, 0, 0)
    result = np.clip(
        image,
        image.min() + dynamic_range * cheat_low,
        image.max() - dynamic_range * cheat_high,
    )
    if isinstance(image, np.ma.MaskedArray):
        return result
    return result.data


def sinh_scale(array, bounds=(0, 1), stretch=1):
    minimum, maximum = array.min(), array.max()
    scaled = (array - minimum) / (maximum - minimum) - 0.5
    expanded = np.sinh(scaled * stretch)
    return normalize_range(expanded, bounds)


# TODO: would it be better for these functions to use flat mask arrays
#  rather than coordinate grids? (performance is the same, question is
#  interface convenience)
def bilinear_interpolate(
    input_array: np.ndarray,
    output_shape: tuple[int, int],
    y_coords: np.ndarray,
    x_coords: np.ndarray,
) -> np.ndarray:
    """
    upsample a 2D array, gridding its pixels at y-coordinates given in y_coords
    and x-coordinates given in x_coords and linearly interpolating from there,
    first in the x direction and then in the y direction. y_coords and x_coords
    are 2D arrays of the same size like those created by np.meshgrid,
    defining locations in N**2
    This is the general (gridded) case and should work correctly for any
    pixel class defined as a partition of a grid. Use the faster
    bilinear_interpolate_subgrid() for pixel classes that can be defined as
    locations relative to a regular tiling of a grid.
    """
    horizontal = np.zeros(output_shape, dtype=input_array.dtype)
    vertical = np.zeros(output_shape, dtype=input_array.dtype)
    rows = np.unique(y_coords)
    for row_ix in rows:
        coordinate_reference_indices = np.where(y_coords == row_ix)
        column_indices = x_coords[coordinate_reference_indices]
        horizontal[row_ix] = np.interp(
            np.arange(output_shape[1]),
            column_indices,
            input_array[coordinate_reference_indices],
        )
    # column indices are immaterial now, because we've interpolated the rows.
    for column_ix in np.arange(output_shape[1]):
        vertical[:, column_ix] = np.interp(
            np.arange(output_shape[0]), rows, horizontal[rows, column_ix]
        )
    return vertical


def bilinear_interpolate_subgrid(
    rows: np.ndarray,
    columns: np.ndarray,
    input_array: np.ndarray,
    output_shape: tuple[int, int],
) -> np.ndarray:
    """
    interpolate 2D values to a 2D array, gridding those values
    according to a regular pattern. input array must be of an
    integer or float dtype.

    this is a special case for pixel classes that are defined as unique
    positions within m x n subgrids that tile the coordinate space.
    in particular, it will work for any conventional Bayer pattern,
    when each pattern cell is treated as a separate pixel 'class' (even if
    pixels are of the 'same' color), as well as a variety of more complex
    patterns. use the slower bilinear_interpolate() for general gridded-data
    cases.
    """
    horizontal = np.empty(output_shape, dtype=input_array.dtype)
    vertical = np.empty(output_shape, dtype=input_array.dtype)
    # because of the 'subgrid' assumption, in any row that contains pixels,
    # their column indices are the same (and vice versa)
    for row_ix, row in enumerate(rows):
        horizontal[row] = np.interp(
            np.arange(output_shape[1]),
            columns,
            input_array[row_ix, :],
        )
    for output_column in np.arange(output_shape[1]):
        vertical[:, output_column] = np.interp(
            np.arange(output_shape[0]), rows, horizontal[rows, output_column]
        )
    return vertical


def apply_image_filter(image, image_filter=None):
    """
    unpacking / pseudo-dispatch function

    `image_filter` is a dict() that contains 'params' and 'function'
    keys. The optional 'params' value is a parameters dict() and the 'function'
    value is the corresponding function to run on `image`.
    """
    if image_filter is None:
        return image
    return image_filter["function"](image, **image_filter.get("params", {}))


def mapfilter(predicate, key, map_sequence):
    new_sequence = []
    for mapping in map_sequence:
        obj = mapping if key is None else mapping.get(key)
        if predicate(obj):
            new_sequence.append(mapping)
    return new_sequence


def make_mask_passer(func, mask_nans=True):
    def mask_passer(array, *args, **kwargs):
        transformed = func(array, *args, **kwargs)
        if isinstance(array, np.ma.MaskedArray):
            mask = array.mask
            if mask_nans:
                mask[np.isnan(transformed)] = True
            return np.ma.MaskedArray(transformed, mask)
        return transformed

    return mask_passer


def get_all_bands(instruction: Mapping):
    """
    helper function for look set analysis: get all bands mentioned in an
    instructions, including an instructions with nested bands
    """
    return chain.from_iterable(dig_for_values(instruction, "bands"))


def get_all_bands_from_all(instructions: Collection[Mapping]):
    """
    helper function for look set analysis: get all bands mentioned in all
    instructions, including from instructions with nested bands
    """
    return set(chain.from_iterable(map(get_all_bands, instructions)))


def closest_ratio(integer, target):
    from sympy import factorint

    factors = tuple(
        chain.from_iterable(
            [[f] * count for f, count in factorint(integer).items()]
        )
    )
    factors = tuple(reversed(factors))
    factorizations = []
    possibilities = []
    for unique in set(factors):
        left = [unique]
        right = [f for f in factors if f not in left]
        quantity = factors.count(unique)
        right += [unique] * (quantity - 1)
        ratio = target
        new_left, new_right = left, right
        while ratio <= target:
            for check_ix in range(len(right)):
                new_left = left + [right[check_ix]]
                for f in new_left:
                    if f in new_right:
                        intermediate = [r for r in right if r != f]
                        quantity = right.count(f)
                        intermediate += [f] * (quantity - 1)
                        new_right = intermediate
                ratio = reduce(mul, new_left) / reduce(mul, new_right)
            if ratio <= target:
                left, right = new_left, new_right
        possibilities.append(reduce(mul, left) / reduce(mul, right))
        factorizations.append((left, right))
        possibilities.append(ratio)
        factorizations.append((new_left, new_right))
    differences = [abs(p - target) for p in possibilities]
    closest = min(differences)
    for p, diff, fact in zip(possibilities, differences, factorizations):
        if diff == closest:
            return p, fact


def closest_aspect(new_size, aspect_ratio):
    _, factors = closest_ratio(new_size, aspect_ratio)
    left, right = factors
    return reduce(mul, right), reduce(mul, left)


def strict_reshape(array, aspect_ratio):
    return array.reshape(closest_aspect(array.size, aspect_ratio))


def ravel_valid(array, copy=True):
    values = array.ravel()
    if isinstance(values, np.ma.MaskedArray):
        if copy is True:
            values = values.copy()
        values[values.mask] = np.nan
        return values[np.isfinite(values)].data
    return values[np.isfinite(values)]


def setmask(array, value, copy=True):
    if not isinstance(array, np.ma.MaskedArray):
        return array
    if copy is True:
        data = array.data.copy()
    else:
        data = array.data
    data[array.mask] = value
    return data


def zero_mask(array, copy=True):
    return setmask(array, 0, copy)


def nanmask(array, copy=True):
    return setmask(array, np.nan, copy)


def make_mask_canvas(image, unmasked_alpha=1, masked_alpha=0):
    canvas = np.full(image.shape[:2], unmasked_alpha, dtype="float16")
    if len(image.shape) > 2:
        mask = image.mask.sum(axis=-1).astype(bool)
    else:
        mask = image.mask
    canvas[mask] = masked_alpha
    return canvas


def colorfill_maskedarray(
    masked_array: np.ma.MaskedArray,
    color: Union[float, tuple[float, float, float]] = 0.45,
    mask_alpha=None,
    unmasked_alpha=None,
):
    """
    masked_array: 2-D masked array or a 3-D masked array with last axis of
    length 3. should be float normalized to 0-1.
    color: optionally-specified RGB color or single integer for gray
    (default medium gray).
    if both alpha parameters are None, return a 3-D array with masked values
    filled with color.
    otherwise, return a 4-D array (last axis intended as an alpha plane),
    with alpha value equal to unmasked_alpha outside the mask and masked_alpha
    inside the mask (either defaults to 0 if not set)
    """
    if (mask_alpha is None) and (unmasked_alpha is not None):
        mask_alpha = 0
    if (unmasked_alpha is None) and (mask_alpha is not None):
        unmasked_alpha = 0
    if len(masked_array.shape) == 2:
        if isinstance(color, (int, float)):
            color = (color, color, color)
        stack = [masked_array.filled(color[ix]) for ix in range(3)]
        if mask_alpha is not None:
            stack.append(
                make_mask_canvas(masked_array, unmasked_alpha, mask_alpha)
            )
        return np.dstack(stack)
    if masked_array.shape[-1] != 3:
        raise ValueError("3-D arrays must have last axis of length = 3")
    if isinstance(color, (int, float)):
        stack = masked_array.filled(color)
    else:
        stack = np.dstack(
            [masked_array[:, :, ix].filled(color[ix]) for ix in range(3)]
        )
    if mask_alpha is None:
        return stack
    return np.dstack(
        [stack, make_mask_canvas(stack, unmasked_alpha, mask_alpha)]
    )


def maskwhere(images, constants):
    return reduce(np.logical_or, [np.isin(i, constants) for i in images])


def pick_mask_constructors(region):
    if region == "inner":
        return np.ma.masked_outside, np.logical_or
    elif region == "outer":
        return np.ma.masked_inside, np.logical_and
    raise ValueError(f"region={region}; region must be 'inner' or 'outer'")


def centered_indices(array):
    y, x = np.indices(array.shape)
    y0, x0 = (array.shape[0] - 1) / 2, (array.shape[1] - 1) / 2
    return y - y0, x - x0


def radial_index(array):
    y_ix, x_ix = centered_indices(array)
    return np.sqrt(y_ix**2 + x_ix**2)


def join_cut_mask(array, cut_mask, copy=True):
    if isinstance(array, np.ma.MaskedArray):
        if copy is True:
            array = array.copy()
        array.mask = np.logical_or(cut_mask, array.mask)
    else:
        array = np.ma.MaskedArray(array, mask=cut_mask)
    return array


def cut_annulus(array, bounds, region="inner", copy=True):
    mask_method, _ = pick_mask_constructors(region)
    distance = radial_index(array)
    pass_min, pass_max = bounds
    pass_min = pass_min if pass_min is not None else distance.min()
    pass_max = pass_max if pass_max is not None else distance.max()
    cut_mask = mask_method(distance, pass_min, pass_max).mask
    return join_cut_mask(array, cut_mask, copy)


def cut_rectangle(array, bounds, region="inner", center=True, copy=True):
    mask_method, op = pick_mask_constructors(region)
    if center is True:
        y_dist, x_dist = centered_indices(array)
    else:
        y_dist, x_dist = np.indices(array.shape)
    masks = []
    x_bounds, y_bounds = bounds
    for bound, dist in zip((x_bounds, y_bounds), (x_dist, y_dist)):
        pass_min, pass_max = bound
        pass_min = pass_min if pass_min is not None else dist.min()
        pass_max = pass_max if pass_max is not None else dist.max()
        masks.append(mask_method(dist, pass_min, pass_max).mask)
    cut_mask = op(*masks)
    return join_cut_mask(array, cut_mask, copy)


def zerocut(*args, **kwargs):
    return zero_mask(cut_rectangle(*args, **kwargs))



"""
utility functions for dealing with matplotlib
"""
import io
from math import ceil, sqrt

from dustgoggles.composition import Composition
from matplotlib.colorbar import Colorbar
from matplotlib.axes import Subplot
from matplotlib.figure import Figure
from matplotlib import pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
import PIL.Image

from marslab.imgops.imgutils import std_clip


def get_mpl_image(fig: Figure):
    """
    tries to get the last axis from a mpl figure.
    cranky and fault-intolerant
    """
    # todo: this does not work for the general case. investigate.
    buffer = io.BytesIO()
    fig.savefig(buffer, bbox_inches="tight", pad_inches=0)
    return PIL.Image.open(buffer)


def remove_ticks(ax):
    """remove ticks from axis"""
    if isinstance(ax, Colorbar):
        ax.set_ticks([])
        return
    ax.set_xticks([])
    ax.set_yticks([])


def despine(ax: Subplot, edges=("top", "bottom", "left", "right")):
    # Remove axes and bounding box for a given subplot object axes
    if isinstance(ax, Colorbar):
        ax.outline.set_visible(False)
        return
    for p in edges:
        ax.spines[p].set_visible(False)


def strip_axes(ax: Subplot):
    remove_ticks(ax)
    despine(ax)


def set_colorbar_font(colorbar, colorbar_fp):
    """set font of colorbar"""
    for tick in colorbar.ax.get_yticklabels():
        tick.set_font_properties(colorbar_fp)
    colorbar.ax.get_yaxis().get_offset_text().set_font_properties(colorbar_fp)


def set_label(
    fig: Figure, text, ax_ix=0, fontproperties=None, loc="center", x_or_y="x"
):
    """
    convenience wrapper for mpl.axes._subplots.AxesSubplot.xlabel / ylabel
    """
    ax = fig.axes[ax_ix]
    if x_or_y == "x":
        method = ax.set_xlabel
    elif x_or_y == "y":
        method = ax.set_ylabel
    else:
        raise ValueError('x_or_y should be "x" or "y"')
    return method(text, loc=loc, fontproperties=fontproperties)


def attach_axis(ax: Subplot = None, where="right", size="50%", pad=0.1):
    if ax is None:
        ax = plt.gca()
    divider = make_axes_locatable(ax)
    return divider.append_axes(where, size=size, pad=pad)


def prefigure(func, title=None):
    def prefigured(*args, **kwargs):
        plt.figure()
        result = func(*args, **kwargs)
        if title is not None:
            plt.title(title)
        return result
    return prefigured


def clipshow_factory(title=None):
    return Composition(
        {'clip': std_clip, 'plot': prefigure(plt.imshow, title)},
        inserts={'clip': {'sigma': 1}}
    )


def gridshow(arrays, kwarg_seq=None, *imshow_args, strip=True, **imshow_kwargs):
    size = ceil(sqrt(len(arrays)))
    fig, grid = plt.subplots(size, size)
    kwarg_seq = [] if kwarg_seq is None else kwarg_seq
    kwarg_seq = [
        {} if i + 1 > len(kwarg_seq) else kwarg_seq[i] for i in range(len(arrays))
    ]
    for cell_ix in range(len(arrays)):
        x, y = (cell_ix % size, (cell_ix // size) % size)
        cell = grid[x][y]
        cell.imshow(
            arrays[cell_ix], *imshow_args, **kwarg_seq[cell_ix], **imshow_kwargs
        )
    if strip is True:
        tuple(map(strip_axes, grid.ravel()))
    fig.tight_layout()
    return fig, grid


def dpi_from_image(fig, ax_ix=0, im_ix=0):
    image = fig.axes[ax_ix].get_images()[im_ix]
    _, im_x_pix = image.get_size()
    dpi_transform = fig.dpi_scale_trans.inverted()
    im_x_in = image.get_window_extent().transformed(dpi_transform).bounds[2]
    return im_x_pix / im_x_in



from marslab.imgops import debayer
import numpy as np


class TestMakePatternMasks:

    def test_make_pattern_masks(self):
        rggb = debayer.make_pattern_masks((4, 4), debayer.RGGB_PATTERN, (2, 2))
        assert (rggb['red'][0] == np.array([0, 2, 0, 2])).all()
        assert (rggb['blue'][1] == np.array([1, 1, 3, 3])).all()




"""
BandSet subclass for processing Mastcam-Z IOFs.

Note that this is closely related to asdf.zcam_bandset.ZcamBandset, but omits
many features of that implementation that only function in the context of the
full asdf workflow.
"""
import re
from functools import partial
from pathlib import Path

import pandas as pd
import pdr
from cytoolz import juxt

from marslab.bandset import BandSet
from marslab.compat.xcam import (
    DERIVED_CAM_DICT, BAND_TO_BAYER,
)
from marslab.imgops.debayer import RGGB_PATTERN
from marslab.imgops.loaders import pdr_load
import marslab.parse as mp

ZCAM_FN_PARSERS = {
    "SOL": mp.sol,
    "SITE": mp.site,
    "DRIVE": mp.drive,
    "SEQ_ID": mp.sequence,
    "CTIME": mp.secondary_timestamp,
    "ZOOM": mp.cam_specific,
    "FILTER": mp.color_filter,
    "PRODUCT_TYPE": mp.product_type,
    "THUMBNAIL": mp.thumbnail,
}


def parse_zcam_fn(path):
    """use mp.parse rules to get basic file identifiers"""
    filename = Path(path).name
    try:
        values = list(juxt(*ZCAM_FN_PARSERS.values())(filename))
        # chop off currently not-used-as-specified stereo counter
        values[5] = values[5][1:]
        # just keep filter name, not SIS-nominal wavelength
        values[6] = values[6][:2]
        parsed = {
            field: value
            for field, value in zip(ZCAM_FN_PARSERS.keys(), values)
        }
        parsed["PATH"] = str(path)
        return parsed
    except (KeyError, IndexError, ValueError):
        return None


def ls_zcam(root_dir):
    files = [file for file in Path(root_dir).iterdir()]
    matches = [
        file
        for file in files
        if not re.match(r".*\.(xml|lbl)$", str(file), re.I)
    ]
    products = tuple(filter(None, map(parse_zcam_fn, matches)))
    if products:
        return (
            pd.DataFrame(products)
            .sort_values(by="CTIME")
            .reset_index(drop=True)
        )
    return None


def setup_zcam_bandset_metadata(metadata):
    if "FILTER" in metadata.columns:
        metadata["BAND"] = metadata["FILTER"]
        metadata.drop("FILTER", axis=1)
    metadata.index = metadata["BAND"]
    # add references to the secret bands hidden inside the L0 and R0 images
    bayer_filter_rows = []
    for eye in ("L", "R"):
        if eye + "0" not in metadata.index:
            continue
        eye_row = metadata.loc[eye + "0"]
        for color in ("R", "G", "B"):
            eye_color_row = eye_row.copy()
            eye_color_row["BAND"] = eye + "0" + color
            eye_color_row.name = eye + "0" + color
            bayer_filter_rows.append(eye_color_row)
        metadata = metadata.drop(eye_row.name)
    if bayer_filter_rows:
        metadata = pd.concat(
            (metadata, pd.concat(bayer_filter_rows, axis=1).T)
        )
    # add wavelengths and bayer pixel mappings
    metadata["WAVELENGTH"] = pd.Series(DERIVED_CAM_DICT["ZCAM"]["filters"])[
        metadata["BAND"]
    ]
    metadata["BAYER_PIXEL"] = pd.Series(BAND_TO_BAYER["ZCAM"])[
        metadata["BAND"]
    ]
    return metadata.reset_index(drop=True)


class ZcamBandSet(BandSet):
    def __init__(self, files):
        load_method = partial(pdr_load, preserve_constants=[0])
        bayer_info = {"pattern": RGGB_PATTERN}
        super().__init__(
            metadata=files, load_method=load_method, bayer_info=bayer_info
        )
        self.metadata = setup_zcam_bandset_metadata(files)
        for path in self.metadata["PATH"].unique():
            self.precached[path] = pdr.Data(
                path, label_fn=path, skip_existence_check=True
            )
        self.check_onboard_debayer(fix_metadata=True)

    def check_onboard_debayer(self, *, fix_metadata=False):
        """
        if this observation was debayered onboard, drop all references to
        bayer pixels to avoid inappropriate debayering later. add
        references to additional bands of R0 / L0 image files (that don't
        exist in raw bayer images).
        """
        if tuple(self.precached.values())[0].metaget(
                'BAYER_METHOD') == 'RAW_BAYER':
            return False
        if fix_metadata:
            self.metadata["BAYER_PIXEL"] = None
            for color, ix in zip(("R", "G", "B"), (0, 1, 2)):
                self.metadata.loc[
                    self.metadata["BAND"].str.endswith(color), "IX"
                ] = ix
        return True



"""
mappings between filter characteristic wavelengths and designations,
along with a bunch of derived related values for "XCAM" 
instruments (PCAM, MCAM, ZCAM...), affording consistent interpretation 
of operations on individual spectra
"""
from collections.abc import Mapping, Sequence
from itertools import chain, combinations
from math import floor
from statistics import mean
from types import MappingProxyType
from typing import Optional

import numpy as np
import pandas as pd
import pandas.api.types
from cytoolz import merge, valfilter
from dustgoggles.pivot import split_on
from marslab.imgops.regions import count_rois_on_image
from more_itertools import windowed

WAVELENGTH_TO_FILTER = {
    "CCAM": {
        400: "400",
        440: "440",
        535: "535",
        600: "600",
        670: "670",
        750: "750",
        840: "840",
    },
    "ZCAM": {
        "L": {
            630: "L0R",
            544: "L0G",
            480: "L0B",
            800: "L1",
            754: "L2",
            677: "L3",
            605: "L4",
            528: "L5",
            442: "L6",
        },
        "R": {
            631: "R0R",
            544: "R0G",
            480: "R0B",
            800: "R1",
            866: "R2",
            910: "R3",
            939: "R4",
            978: "R5",
            1022: "R6",
        },
    },
    "MCAM": {
        "L": {
            482: "L0B",  #
            493: "L0B",  # Accepted value of L0B has changed over time
            495: "L0B",  #
            554: "L0G",
            640: "L0R",
            527: "L1",
            445: "L2",
            751: "L3",
            676: "L4",
            867: "L5",
            1012: "L6",
        },
        "R": {
            482: "R0B",  #
            493: "R0B",  # Accepted value of R0B has changed over time
            495: "R0B",  #
            551: "R0G",
            638: "R0R",
            527: "R1",
            447: "R2",  #
            805: "R3",
            908: "R4",
            937: "R5",
            1013: "R6",  #
        },
    },
}

# Scale to the effective resolution or bayer element ratio of each band
FILTER_TO_RESOLUTION_FACTOR = {
    "ZCAM": {
        "L1": 1,
        "L2": 1,
        "L3": 1,
        "L4": 1,
        "L5": 2,
        "L6": 1,
        "R1": 1,
        "R2": 4,
        "R3": 4,
        "R4": 4,
        "R5": 4,
        "R6": 4,
        "L0G": 2,
        "L0B": 1,
        "L0R": 1,
        "R0G": 2,
        "R0B": 1,
        "R0R": 1,
    }
}

# rules currently in use:
# set of virtual filters === the set of pairs of real filters with nominal
# band centers within 5 nm of one another
# the virtual mean reflectance in an ROI for a virtual filter is the
# arithmetic mean of the mean reflectance values in that ROI for the two real
# filters in its associated pair.
# the nominal band center of a virtual filter is the arithmetic mean of the
# nominal band centers of the two real filters in its associated pair.


def make_xcam_filter_dict(abbreviation):
    """
    form filter: wavelength dictionary for mastcam-family instruments
    """
    if abbreviation == "CCAM":
        return {
            name: wavelength
            for wavelength, name in sorted(
                WAVELENGTH_TO_FILTER[abbreviation].items(),
                key=lambda item: item[1],
            )
        }
    left = {
        name: wavelength
        for wavelength, name in WAVELENGTH_TO_FILTER[abbreviation]["L"].items()
    }
    right = {
        name: wavelength
        for wavelength, name in WAVELENGTH_TO_FILTER[abbreviation]["R"].items()
    }
    return {
        name: wavelength
        for name, wavelength in sorted(
            {**left, **right}.items(), key=lambda item: item[1]
        )
    }


def make_xcam_filter_pairs(abbreviation):
    """
    form list of pairs of close filters for mastcam-family instruments
    """
    filter_dict = make_xcam_filter_dict(abbreviation)
    pairs = []
    for filter_1, filter_2 in combinations(filter_dict, 2):
        if abs(filter_dict[filter_1] - filter_dict[filter_2]) <= 5:
            pairs.append(tuple(sorted([filter_1, filter_2])))
    return tuple(pairs)


def make_virtual_filters(abbreviation):
    """
    form mapping from close filter names to wavelengths for mastcam-family
    """
    filter_dict = make_xcam_filter_dict(abbreviation)
    filter_pairs = make_xcam_filter_pairs(abbreviation)
    return {
        pair[0]
        + "_"
        + pair[1]: floor(mean([filter_dict[pair[0]], filter_dict[pair[1]]]))
        for pair in filter_pairs
    }


def make_virtual_filter_mapping(abbreviation):
    """
    form mapping from close filter names to filter pairs for mastcam-family
    """
    return {
        pair[0] + "_" + pair[1]: pair
        for pair in make_xcam_filter_pairs(abbreviation)
    }


def make_canonical_averaged_filters(abbreviation):
    filter_dict = make_xcam_filter_dict(abbreviation)
    virtual_filters = make_virtual_filters(abbreviation)
    virtual_filter_mapping = make_virtual_filter_mapping(abbreviation)
    retained_filters = {
        filt: filter_dict[filt]
        for filt in filter_dict
        if filt not in chain.from_iterable(virtual_filter_mapping.values())
    }
    caf = {**retained_filters, **virtual_filters}
    return {filt: caf[filt] for filt in sorted(caf, key=lambda x: caf[x])}


XCAM_ABBREVIATIONS = ["MCAM", "ZCAM", "CCAM"]
DERIVED_CAM_DICT = {
    abbrev: {
        "filters": make_xcam_filter_dict(abbrev),
        "virtual_filters": make_virtual_filters(abbrev),
        "virtual_filter_mapping": make_virtual_filter_mapping(abbrev),
        "canonical_averaged_filters": make_canonical_averaged_filters(abbrev),
    }
    for abbrev in XCAM_ABBREVIATIONS
}


def polish_xcam_spectrum(
    spectrum: Mapping[str, float],
    cam_info: Mapping[str, dict],
    scale_to: Optional[Sequence[str, str]] = None,
    average_filters: bool = True,
):
    """
    scale and merge values of a spectrum according to MERSPECT-style rules
    scale_to: None or tuple of (lefteye filter name, righteye filter name)
    """
    values = {}
    lefteye_scale = 1
    righteye_scale = 1
    # don't scale eyes to a value that doesn't exist or if you're asked not to
    if scale_to not in [None, "None"]:
        if all([spectrum.get(comp) for comp in scale_to]):
            scales = (spectrum[scale_to[0]], spectrum[scale_to[1]])
            filter_mean = mean(scales)
            lefteye_scale = filter_mean / scales[0]
            righteye_scale = filter_mean / scales[1]

    real_filters_to_use = list(cam_info["filters"].keys())
    # TODO: get rid of legacy _ERR fields
    if average_filters is True:
        # construct dictionary of averaged filter values
        for v_filter, comps in cam_info["virtual_filter_mapping"].items():
            # do not attempt to average filters if both filters of
            # a pair are not present
            if not all([spectrum.get(comp) for comp in comps]):
                continue
            [real_filters_to_use.remove(comp) for comp in comps]
            values[v_filter] = {
                "wave": cam_info["virtual_filters"][v_filter],
                "mean": mean(
                    (
                        spectrum[comps[0]] * lefteye_scale,
                        spectrum[comps[1]] * righteye_scale,
                    ),
                ),
            }
            if all([comp + "_STD" in spectrum.keys() for comp in comps]):
                values[v_filter]["std"] = (
                    spectrum[comps[0] + "_STD"] ** 2
                    + spectrum[comps[1] + "_STD"] ** 2
                ) ** 0.5
            if all([comp + "_ERR" in spectrum.keys() for comp in comps]):
                values[v_filter]["iof_err"] = (
                    spectrum[comps[0] + "_ERR"] ** 2
                    + spectrum[comps[1] + "_ERR"] ** 2
                ) ** 0.5
    # construct dictionary of leftover real filter values
    for real_filter in real_filters_to_use:
        mean_value = spectrum.get(real_filter)
        if mean_value is None:
            continue
        if real_filter.lower().startswith("r"):
            eye_scale = righteye_scale
        else:
            eye_scale = lefteye_scale
        values[real_filter] = {
            "wave": cam_info["filters"][real_filter],
            "mean": spectrum[real_filter] * eye_scale,
        }
        if real_filter + "_STD" in spectrum.keys():
            values[real_filter]["std"] = (
                spectrum[real_filter + "_STD"] * eye_scale
            )
        if real_filter + "_ERR" in spectrum.keys():
            values[real_filter]["iof_err"] = (
                spectrum[real_filter + "_ERR"] * eye_scale
            )

    return dict(sorted(values.items(), key=lambda item: item[1]["wave"]))


INSTRUMENT_UNCERTAINTIES = {
    # table 7, hayes et al. 2021 https://doi.org/10.1007/s11214-021-00795-x
    # "ZCAM": {
    #     "L0R": 3.3,
    #     "L0G": 3.3,
    #     "L0B": 3.7,
    #     "L1": 1.4,
    #     "L2": 1.1,
    #     "L3": 0.2,
    #     "L4": 1.8,
    #     "L5": 1.6,
    #     "L6": 0.4,
    #     "R0R": 3.7,
    #     "R0G": 4.1,
    #     "R0B": 4.6,
    #     "R1": 0.4,
    #     "R2": 0.3,
    #     "R3": 0.6,
    #     "R4": 0.5,
    #     "R5": 0.8,
    #     "R6": 0.4,
    # },
    # hayes, p. comm, April 2021: use 3% for everything but the
    # bayers for now
    "ZCAM": {
        "L0R": 3.3,
        "L0G": 3.3,
        "L0B": 3.7,
        "L1": 3,
        "L2": 3,
        "L3": 3,
        "L4": 3,
        "L5": 3,
        "L6": 3,
        "R0R": 3.7,
        "R0G": 4.1,
        "R0B": 4.6,
        "R1": 3,
        "R2": 3,
        "R3": 3,
        "R4": 3,
        "R5": 3,
        "R6": 3,
    },
    # table 2, bell et al. 2017 https://doi.org/10.1002/2016EA000219
    "MCAM": {
        "L0R": 1.2,
        "L0G": 0.3,
        "L0B": 5.7,
        "L1": 4.3,
        "L2": 51.0,
        "L3": 0.3,
        "L4": 0.1,
        "L5": 0.3,
        "L6": 1.0,
        "R0R": 1.9,
        "R0G": 1.5,
        "R0B": 2.5,
        "R1": 3.7,
        "R2": 24.5,
        "R3": 3.4,
        "R4": 0.4,
        "R5": 0.5,
        "R6": 1.0,
    },
}

# see M20 Camera SIS
ZCAM_ZOOM_MOTOR_COUNT_TO_FOCAL_LENGTH = {
    0: 26,
    2448: 34,
    3834: 48,
    5196: 63,
    6720: 79,
    8652: 100,
    9600: 110,
}


def piecewise_interpolate_focal_length(zmc):
    z_values = ZCAM_ZOOM_MOTOR_COUNT_TO_FOCAL_LENGTH.keys()
    f_values = ZCAM_ZOOM_MOTOR_COUNT_TO_FOCAL_LENGTH.values()
    if zmc in z_values:
        return ZCAM_ZOOM_MOTOR_COUNT_TO_FOCAL_LENGTH[zmc]
    try:
        for z1z2, f1f2 in zip(windowed(z_values, 2), windowed(f_values, 2)):
            z1, z2 = z1z2
            f1, f2 = f1f2
            if not ((z1 < zmc) and (zmc < z2)):
                continue
            return round((f2 - f1) / (z2 - z1) * zmc + f1, 1)
    except StopIteration:
        raise ValueError(
            str(zmc) + " is outside the range of zoom"
            " motor counts I know how to deal with."
        )


BAND_TO_BAYER = {
    "ZCAM": {
        "L1": "red",
        "L2": "red",
        "L3": "red",
        "L4": "red",
        "L5": ("green_1", "green_2"),
        "L6": "blue",
        "R1": "red",
        "R2": None,
        "R3": None,
        "R4": None,
        "R5": None,
        "R6": None,
        "L0G": ("green_1", "green_2"),
        "L0B": "blue",
        "L0R": "red",
        "R0G": ("green_1", "green_2"),
        "R0B": "blue",
        "R0R": "red",
    },
    "MCAM": {
        "L1": ("green_1", "green_2"),
        "L2": "blue",
        "L3": "red",
        "L4": "red",
        "L5": None,
        "L6": None,
        "R1": ("green_1", "green_2"),
        "R2": "blue",
        "R3": "red",
        "R4": None,
        "R5": None,
        "R6": None,
        # these are apparently debayered onboard or always debayered
        # by the time we see reduced images
        "L0G": None,
        "L0B": None,
        "L0R": None,
        "R0G": None,
        "R0B": None,
        "R0R": None,
    },
}

# TODO: de-vendor downstream (moved up to dustgoggles)
def integerize(df):
    for column in numeric_columns(df):
        if pd.api.types.is_integer_dtype(df[column].dtype):
            continue
        isna, notna = split_on(df[column], df[column].isna())
        if not (notna.round() == notna).all():
            continue
        df.loc[isna.index, column] = ""
        df.loc[notna.index, column] = notna.map("{:.0f}".format)
    return df


def numeric_columns(data: pd.DataFrame) -> list[str]:
    return [
        col
        for col in data.columns
        if pandas.api.types.is_numeric_dtype(data[col])
    ]


def filter_arrays_from(records):
    filtered = map(
        lambda rec: valfilter(lambda v: not isinstance(v, np.ndarray), rec),
        records,
    )
    return list(filtered)


def squeeze_records(records, on="COLOR"):
    records = {
        value: merge([rec for rec in records if rec[on] == value])
        for value in set(rec[on] for rec in records)
    }
    return pd.DataFrame.from_dict(records, "index")


def count_rois_on_xcam_images(
    roi_hdulist: list,
    xcam_image_dict: dict,
    instrument: str,
    pixel_map_dict=None,
    error_map_dict=None,
    bayer_pixel_dict=None,
    special_constants=tuple([0]),
):
    """
    takes an roi hdulist, a dict of xcam images, and returns a marslab data
    section dataframe.

    there are so many potential special cases here that utterly transform
    control flow that we've chosen to structure it differently from the
    quick imaging functions. perhaps this is wrong, though.

    """
    from marslab.imgops.debayer import RGGB_PATTERN, make_bayer

    # unrolling for easier iteration
    roi_hdus = [roi_hdulist[hdu_ix] for hdu_ix in roi_hdulist]
    rois = {}
    for eye in ("LEFT", "RIGHT"):
        hdus = [hdu for hdu in roi_hdus if hdu.header["EYE"].upper() == eye]
        rois[eye] = {hdu.header["NAME"]: hdu.data for hdu in hdus}

    if bayer_pixel_dict is None:
        bayer_pixel_dict = BAND_TO_BAYER[instrument]
    # don't attempt to apply Bayer masks if Bayer pixels are explicitly
    # assigned as None. Permits overriding Bayer pixel selection for use cases
    # like images that arrived at our pipeline already debayered.
    if not all([pixel is None for pixel in bayer_pixel_dict.values()]):
        bayer_masks = make_bayer(
            list(xcam_image_dict.values())[0].shape, RGGB_PATTERN
        )
    else:
        bayer_masks = None
    roi_records = []
    for filter_name in DERIVED_CAM_DICT[instrument]["filters"].keys():
        if (image := xcam_image_dict.get(filter_name)) is None:
            continue
        # bayer-counting logic
        detector_mask = make_detector_mask(
            bayer_masks, bayer_pixel_dict, filter_name, image
        )
        # forbidding saturated and otherwise bad pixels
        detector_mask = apply_pixel_map(
            detector_mask, filter_name, image, pixel_map_dict
        )
        roi_records += count_image(
            image,
            filter_name,
            rois,
            detector_mask,
            error_map_dict,
            FILTER_TO_RESOLUTION_FACTOR.get(instrument),
            special_constants,
        )
    cube_records = aggregate_eye_stats(roi_records, rois)
    roi_records = squeeze_records(filter_arrays_from(roi_records)).drop(
        columns="COLOR"
    )
    base_df = (
        pd.concat([cube_records, roi_records], axis=1)
        .copy()
        .reset_index(drop=True)
    )
    base_df = add_position_metrics(base_df)
    downcast = base_df[numeric_columns(base_df)].astype(np.float32)
    base_df[numeric_columns(base_df)] = downcast.values
    base_df = base_df.copy()
    # enter nan columns for err and mean only -- this is a format
    # standardization choice
    for filter_name in DERIVED_CAM_DICT[instrument]["filters"].keys():
        if filter_name not in base_df.columns:
            base_df[filter_name] = np.nan
            base_df[filter_name + "_STD"] = np.nan
            base_df[filter_name + "_ERR"] = np.nan
    return base_df.copy()


def add_position_metrics(base_df):
    measures = ("ROW", "COLUMN", "DET_RAD", "DET_THETA")
    for measure in measures:
        base_df[measure] = np.nan
    for ix, row in base_df.iterrows():
        existing = row.dropna()
        for measure in measures:
            oculars = [
                ocular
                for ocular in (f"LEFT_{measure}", f"RIGHT_{measure}")
                if ocular in existing.index
            ]
            base_df.loc[ix, measure] = np.mean(
                [row[ocular] for ocular in oculars]
            )
    return base_df


def apply_pixel_map(detector_mask, filter_name, image, pixel_map_dict):
    if pixel_map_dict:
        if filter_name[1] == "0":
            base_pixel_map = pixel_map_dict.get(filter_name[0:2])
        else:
            base_pixel_map = pixel_map_dict.get(filter_name)
        if base_pixel_map is not None:
            # masking bad, no-signal, and saturated pixels
            flag_mask = np.full(image.shape, True)
            flag_mask[np.where(np.isin(base_pixel_map, [1, 2, 4]))] = False
            detector_mask = np.logical_and(detector_mask, flag_mask)
    return detector_mask


def make_detector_mask(bayer_masks, bayer_pixel_dict, filter_name, image):
    if bayer_pixel_dict[filter_name] is not None:
        detector_mask = np.full(image.shape, False)
        bayer_pixels = bayer_pixel_dict[filter_name]
        if isinstance(bayer_pixels, str):
            bayer_pixels = [bayer_pixels]
        for pixel in bayer_pixels:
            bayer_coords = bayer_masks[pixel]
            detector_mask[bayer_coords] = True
    else:
        detector_mask = np.full(image.shape, True)
    return detector_mask


def count_image(
    image,
    filter_name,
    rois,
    detector_mask,
    error_map_dict,
    resolution_dict,
    special_constants,
):
    eye = "LEFT" if filter_name.upper().startswith("L") else "RIGHT"
    roi_counts = count_rois_on_image(
        rois[eye].values(),
        rois[eye].keys(),
        image,
        detector_mask,
        special_constants,
    )
    if error_map_dict:
        ioe = np.array(error_map_dict.get(filter_name), dtype="float64")
        error_image = ioe ** 2 - image
    else:
        error_image = None
    roi_records = []
    for roi_name, counts in roi_counts.items():
        if error_image is not None:
            error = calculate_roi_error(
                counts,
                detector_mask,
                error_image,
                # TODO: onboard-debayered images
                resolution_dict[filter_name],
                rois[eye][roi_name],
                special_constants,
            )
        else:
            error = None
        constant_stats = {
            "COLOR": roi_name,
            filter_name: counts["mean"],
            filter_name + "_STD": counts["var"],
            filter_name + "_MODE": counts["mode"],
            filter_name + "_ERR": error,
        }
        variable_stats = {
            filter_name + "_" + stat.upper(): counts[stat]
            for stat in counts.keys()
        }
        roi_records.append(constant_stats | variable_stats)
    return roi_records


# TODO: This is the wrong calculation for the error; it's just a
#  stand-in until we clarify
def calculate_roi_error(
    counts,
    detector_mask,
    error_image,
    resolution_factor,
    roi,
    special_constants,
):
    from marslab.imgops.regions import count_rois_on_image

    scaled_counts = counts["count"] / resolution_factor
    errors = count_rois_on_image(
        [roi],
        [""],
        error_image + counts["mean"] / scaled_counts,
        detector_mask,
        special_constants,
    )
    return tuple(errors.values())[0]["mean"]


def aggregate_eye_stats(roi_records, rois):
    roi_frame = pd.DataFrame(roi_records)
    for column in numeric_columns(roi_frame):
        roi_frame[column] = roi_frame[column].astype(np.float32)
    cube_records = []
    for eye in ("LEFT", "RIGHT"):
        cube_records += aggregate_single_eye_stats(roi_frame, eye, rois)
    return squeeze_records(filter_arrays_from(cube_records))


def aggregate_single_eye_stats(statframe, eye, rois):
    eye_values = statframe.loc[
        :, statframe.columns.str.match(f"{eye[0].upper()}.*VALUES.*")
    ]
    if len(eye_values.columns) == 0:
        return []
    eye_values.index = statframe["COLOR"]
    melted = pd.melt(eye_values, ignore_index=False).dropna()
    eyestats = [
        aggregate_across_filters(eye, melted, roi_name, rois)
        for roi_name in rois[eye].keys()
    ]
    return eyestats


def aggregate_across_filters(eye, melted, roi_name, rois):
    from marslab.imgops.regions import roi_stats, roi_position

    roi = melted.loc[roi_name]["value"]
    if isinstance(roi, pd.Series):
        roi = np.hstack(roi.to_numpy())
    # this performs stats twice (here and in the per-filter counting) in the
    # degenerate case of ROIs drawn on only one filter, but this is not really
    # a big deal.
    counts, position = roi_stats(roi), roi_position(rois[eye][roi_name])
    base_aggregate_stat = {
        "COLOR": roi_name,
        eye: counts["mean"],
        f"{eye}_ROW": position["y"],
        f"{eye}_COLUMN": position["x"],
        f"{eye}_DET_RAD": position["r"],
        f"{eye}_DET_THETA": position["theta"],
    }
    variable_aggregate_stat = {
        f"{eye}_{stat.upper()}": counts[stat] for stat in counts.keys()
    }
    return base_aggregate_stat | variable_aggregate_stat


# standard translations between eye codes and names
EYE_TERMS = MappingProxyType(
    {"left": "L", "l": "left", "r": "right", "right": "R"}
)

# noinspection PyTypeChecker
REVERSE_EYE_TERMS = MappingProxyType(
    {k: v for k, v in zip(reversed(EYE_TERMS.keys()), EYE_TERMS.values())}
)


def eye_name(string, swap=False):
    terms = EYE_TERMS if swap is False else REVERSE_EYE_TERMS
    try:
        return terms[string.lower()]
    except KeyError:
        raise ValueError("This axis only has left/L and right/R directions.")


def construct_field_ordering(filters, fields):
    initial = (
        "NAME",
        "COLOR",
        "ANALYSIS_NAME",
        "SOL",
        "SEQ_ID",
        "FEATURE",
        "FEATURE_SUBTYPE",
        "DESCRIPTION",
        "SITE",
        "DRIVE",
        "RSM",
        "LTST",
        "INCIDENCE_ANGLE",
        "EMISSION_ANGLE",
        "PHASE_ANGLE",
        "SOLAR_ELEVATION",
        "SOLAR_AZIMUTH",
        "LAT",
        "LON",
        "ODOMETRY",
        "ROVER_ELEVATION",
        "TARGET_ELEVATION",
        "INSTRUMENT",
        "SCLK",
        "UNITS",
    )
    order = []
    for predecessor in initial + filters:
        if predecessor in fields:
            order.append(predecessor)
    stats = map(
        lambda f: str(f).replace(f"{filters[0]}_", ""),
        filter(lambda f: str(f).startswith(f"{filters[0]}_"), fields),
    )
    if "STD" in stats:
        stats = ["STD"] + [s for s in stats if stats != "STD"]
    for stat in stats:
        order += list(map(lambda s: f"{s}_{stat}", filters))
    order += [f for f in fields if f not in order]
    return order



from marslab import spectops
import numpy as np
import pandas as pd


class TestSpectopsErrorCalculations:
    def test_errors_in_quadrature_digit(self):
        assert spectops.addition_in_quadrature([2]) == 2

    def test_errors_in_quadrature_tuple(self):
        assert spectops.addition_in_quadrature([2, 2]) == np.sqrt(2 * 2 ** 2)

    def test_errors_in_quadrature_lists(self):
        assert spectops.addition_in_quadrature([[1, 2], [3, 4]])[0] == np.sqrt(
            1 ** 2 + 3 ** 2
        )

    def test_errors_in_quadrature_ndarray(self):
        assert spectops.addition_in_quadrature(np.array([[1, 2], [3, 4]]))[
            0
        ] == np.sqrt(1 ** 2 + 3 ** 2)

    def test_errors_in_quadrature_dataframe(self):
        assert spectops.addition_in_quadrature(pd.DataFrame([[1, 2], [3, 4]]))[
            0
        ] == np.sqrt(1 ** 2 + 3 ** 2)


class TestSpectopsRatio:
    def test_ratio_1_list(self):
        assert spectops.ratio([10, 20])[0] == 0.5

    def test_ratio_1_ndarray(self):
        assert spectops.ratio(np.array([10, 20]))[0] == 0.5

    # def test_ratio_sequence_1_series(self):
    #    assert spectops.ratio(pd.Series([10,20]))[0] == 0.5

    def test_ratio_2_list(self):
        assert spectops.ratio([[5, 20], [10, 40]])[0][0] == 0.5
        assert spectops.ratio([[5, 20], [10, 40]])[0][1] == 0.5

    def test_ratio_2_ndarray(self):
        assert spectops.ratio(np.array([[5, 20], [10, 40]]))[0][0] == 0.5
        assert spectops.ratio(np.array([[5, 20], [10, 40]]))[0][1] == 0.5

    # def test_ratio_sequence_2_dataframe(self):
    #    assert spectops.ratio(pd.DataFrame([[10,20],[30,40]]))[0][0] == 0.5
    #    assert spectops.ratio(pd.DataFrame([[10,20],[30,40]]))[0][1] == 0.75

    def test_ratio_w_errors(self):
        assert spectops.ratio([10, 20], errors=[2, 2])[0] == 0.5
        assert spectops.ratio([10, 20], errors=[2, 2])[1] == np.sqrt(
            2 * 2 ** 2
        )


class TestSpectopsBandAvg:
    def test_band_avg_1(self):
        assert spectops.band_avg([2, 4], errors=[2, 2])[0] == 3
        assert spectops.band_avg([2, 4], errors=[2, 2])[1] == np.sqrt(
            2 * 2 ** 2
        )

    def test_band_avg_2(self):
        assert spectops.band_avg([[2, 2], [3, 3]], errors=[2, 2])[0][0] == 2.5


class TestSpectopsSlope:
    def test_slope_1(self):
        assert (
            spectops.slope([1, 2], wavelengths=[0, 1], errors=[2, 2])[0] == 1
        )
        assert spectops.slope([1, 2], wavelengths=[0, 1], errors=[2, 2])[
            1
        ] == np.sqrt(2 * 2 ** 2)


class TestSpectopsBandDepth:
    def test_band_depth_1(self):
        assert (
            spectops.band_depth(
                [1, 1, 0], wavelengths=[1, 3, 2], errors=[2, 2, 2]
            )[0]
            == 1
        )
        assert spectops.band_depth(
            [1, 1, 0], wavelengths=[1, 3, 2], errors=[2, 2, 2]
        )[1] == np.sqrt(3 * 2 ** 2)


class TestSpectopsBandMin:
    def test_band_min_1(self):
        assert (
            spectops.band_min([1, 2, 3, 4], wavelengths=[1, 2, 3, 4])[0] == 1
        )

    def test_band_min_2(self):
        assert (
            spectops.band_min([[1, 2], [2, 1]], wavelengths=[1, 2])[0][0] == 1
        )
        assert (
            spectops.band_min([[1, 2], [2, 1]], wavelengths=[1, 2])[0][1] == 2
        )


class TestSpectopsBandMax:
    def test_band_min_1(self):
        assert (
            spectops.band_max([1, 2, 3, 4], wavelengths=[1, 2, 3, 4])[0] == 4
        )

    def test_band_max_2(self):
        assert (
            spectops.band_max([[1, 2], [2, 1]], wavelengths=[1, 2])[0][0] == 2
        )
        assert (
            spectops.band_max([[1, 2], [2, 1]], wavelengths=[1, 2])[0][1] == 1
        )



import numpy as np
from scipy.ndimage import sobel, distance_transform_edt
from scipy.stats import skew, kurtosis, mode


def furthest_from_edge(image):
    distances = distance_transform_edt(image)
    return np.unravel_index(distances.argmax(), distances.shape)


def roi_position(roi):
    y, x = tuple(map(np.median, np.nonzero(roi)))
    return {
        "y": y,
        "x": x,
        "r": np.sqrt(y**2 + x**2),
        "theta": np.arctan2(y, x)
    }


def roi_stats(array, extended=True):
    if isinstance(array, np.ma.MaskedArray):
        vals = array.compressed()
    else:
        vals = array.ravel()
    # TODO: why are we naming it "var" here and renaming it "STD" elsewhere
    base = {"mean": vals.mean(), "var": vals.std(), "values": vals}
    if extended:
        base |= {
            "min": vals.min(),
            "max": vals.max(),
            "total": vals.sum(),
            "count": vals.size,
            "skew": skew(vals),
            "kurtosis": kurtosis(vals),
            "mode": mode(vals, keepdims=False).mode,
            "median": np.ma.median(vals),
        }
    return base


def make_roi_edgemaps(roi_fits, calculate_centers=True):
    roi_edgemaps = {}
    for roi_ix in roi_fits:
        roi_array = roi_fits[roi_ix].data
        name = roi_fits[roi_ix].header["EXTNAME"].lower()
        edgemap = make_edgemap(roi_array).astype("uint8")
        roi_edgemaps[name] = {"edgemap": edgemap}
        if calculate_centers:
            center_y, center_x = furthest_from_edge(roi_array)
            roi_edgemaps[name]["center"] = (center_x, center_y)
    return roi_edgemaps


# TODO: nasty hack, consolidate
def draw_edgemaps_on_axis(
    ax,
    edgemap_dict,
    inscribe_names=False,
    fontproperties=None,
    width=4,
    colorize=True,
):
    for roi_name, edgemap in edgemap_dict.items():
        if colorize and ("color" in edgemap.keys()):
            color = edgemap["color"]
        else:
            color = "white"
        edge_y, edge_x = np.nonzero(edgemap["edgemap"])
        ax.scatter(edge_x, edge_y, marker=".", s=width, color=color, alpha=0.4)
        if inscribe_names:
            ax.annotate(
                roi_name,
                (edgemap["center"][0] - 15, edgemap["center"][1]),
                color="white",
                fontproperties=fontproperties,
                alpha=0.8,
            )


def draw_edgemaps_on_image(
    image,
    edgemap_dict,
    inscribe_names=False,
    fontproperties=None,
    width=4,
    colorize=True,
):
    from marslab.imgops.pltutils import strip_axes
    import matplotlib.pyplot as plt

    fig = plt.figure()
    ax = fig.add_subplot()
    ax.imshow(image, interpolation=None)
    for roi_name, edgemap in edgemap_dict.items():
        if colorize and ("color" in edgemap.keys()):
            color = edgemap["color"]
        else:
            color = "white"
        edge_y, edge_x = np.nonzero(edgemap["edgemap"])
        plt.scatter(
            edge_x, edge_y, marker=".", s=width, color=color, alpha=0.4
        )
        if inscribe_names:
            plt.annotate(
                roi_name,
                (edgemap["center"][0] - 15, edgemap["center"][1]),
                color="white",
                fontproperties=fontproperties,
                alpha=0.8,
            )
    strip_axes(ax)
    return fig


def make_edgemap(image, threshold=0.01):
    edge = np.hypot(sobel(image, 0), sobel(image, 1))
    edge[np.where(edge > threshold)] = 1
    return edge / edge.max()


def count_rois_on_image(
    roi_arrays,
    roi_names,
    image,
    detector_mask=None,
    special_constants=None,
):
    """
    expects ROIs 'unpacked' from their FITS format -- this is basically because
    we want applications to be able to prefilter the list without sending a
    bunch of additional information to this function, and also to count arrays
    made in other ways without rolling them into FITS files!
    """
    count_dict = {}
    if special_constants is not None:
        special_mask = np.full(image.shape, True)
        special_mask[np.isin(image, special_constants)] = False
    for roi_mask, roi_name in zip(roi_arrays, roi_names):
        roi_mask = roi_mask.astype(bool)
        assert (
            roi_mask.shape == image.shape
        ), "it seems like this ROI might have been drawn on a different image."
        if detector_mask is not None:
            roi_mask = np.logical_and(roi_mask, detector_mask)
        if special_constants is not None:
            roi_mask = np.logical_and(roi_mask, special_mask)
        # generally cases like: this ROI is drawn entirely within the missing
        # area of a partial
        if np.all(~roi_mask):
            continue
        count_dict[roi_name] = roi_stats(image[roi_mask])
    return count_dict


def make_roi_hdu(input_array, roi_name, metadata_dict):
    """
    make an individual HDU for a marslab .roi file
    """
    from astropy.io import fits

    roi_hdu = fits.PrimaryHDU(input_array)
    # this parameter is separate and mandatory because we really
    # need every ROI to have some kind of distinguishing name
    roi_hdu.header["NAME"] = roi_name
    # not being strict right now
    for field, value in metadata_dict.items():
        roi_hdu.header[field] = value
    roi_hdu.name = roi_name
    return roi_hdu


def select_roi_by_ix(rois: np.ndarray, color_ix: int) -> np.ndarray:
    roi = np.zeros(shape=rois.shape, dtype=rois.dtype)
    roi[np.where(rois == color_ix)] = 1
    return roi



from marslab.compat import xcam

def test_derived_cam_dict_construction():
    derived = xcam.DERIVED_CAM_DICT
    assert max(list(derived['ZCAM']['filters'].values())) == 1022
    assert derived['MCAM']['virtual_filter_mapping']['L2_R2'] == ('L2', 'R2')
    assert min(list(derived['MCAM']['filters'].values())) == 445
    assert 'L0G_R0G' in derived['ZCAM']['virtual_filters']
    assert 'L0R_R0R' in derived['MCAM']['virtual_filters']





"""
reference implementations of elementary spectrum operations.
overloaded to take dataframes, ndarrays, sequences, and scalars.

generic signature: (reflectance, errors, wavelength)

order of band shoulders is: left, right, center --
or unordered for things like average

Caution: everything must be passed in 'last index fastest' / 'row-primary'
orientation. this means that a lot of pandas dataframes will need to be
transposed to get meaningful results.
"""
from typing import Sequence, Union, Optional

import numpy as np
from numpy.linalg import norm
import pandas as pd


# TODO: make all kwarg signatures flatly identical


def preprocess_input(reflectance, errors=None, wavelengths=None):
    """
    did someone pass this function something silly? raise an issue.
    then cast everything to ndarray.
    """
    if errors is not None:
        if isinstance(reflectance, (np.ndarray, pd.DataFrame, pd.Series)):
            assert reflectance.shape == errors.shape
        else:
            assert len(reflectance) == len(errors)
    if wavelengths is not None:
        if isinstance(reflectance, (np.ndarray, pd.DataFrame, pd.Series)):
            assert reflectance.shape[0] == len(wavelengths)
        else:
            assert len(reflectance) == len(wavelengths)
    return [np.asarray(thing) for thing in (reflectance, errors, wavelengths)]


def reindex_or_mask(*object_pairs):
    returning = []
    for pair in object_pairs:
        thing = pair[0]
        if isinstance(pair[1], (pd.DataFrame, pd.Series)):
            thing = pd.Series(thing, index=pair[1].columns)
        elif isinstance(pair[1], Sequence):
            masked = filter(lambda a: isinstance(a, np.ma.MaskedArray), pair[1])
            masks = [m.mask for m in masked]
            if len(masks) > 0:
                thing = np.ma.masked_array(thing, mask=sum(masks))
        returning.append(thing)
    return returning


def addition_in_quadrature(errors):
    """
    take what numpy thinks is the norm.
    in most cases this should be equivalent to
    summing in quadrature. this is just what
    we're doing for now, I guess.
    """
    if None in errors:
        return None
    return norm(errors, axis=0)


def ratio(
    reflectance: Union[pd.DataFrame, pd.Series, Sequence, np.ndarray],
    errors: Union[pd.DataFrame, pd.Series, Sequence, np.ndarray, None] = None,
    _wavelengths: Optional[Sequence] = None,
):
    """
    just a ratio function. notionally, ratio of reflectance
    values between two filters. looks at the first two 'rows'
    or elements of what it's passed...if you pass it more,
    it ignores them. ignores wavelength values.
    """
    reflectance_array, error_array, _wavelength_array = preprocess_input(
        reflectance, errors, _wavelengths
    )
    ratio_value = reflectance_array[0] / reflectance_array[1]
    if None not in error_array:
        error_array = error_array[0:2]
    error_values = addition_in_quadrature(error_array)
    return reindex_or_mask(
        (ratio_value, reflectance), (error_values, errors)
    )


def band_avg(
    reflectance: Union[pd.DataFrame, pd.Series, Sequence, np.ndarray],
    errors: Union[pd.DataFrame, pd.Series, Sequence, np.ndarray, None] = None,
    _wavelengths: Optional[Sequence] = None,
):
    """
    just an averaging function. notionally, average of all filters within a
    contiguous block of wavelengths. your responsibility to make sure
    you didn't skip any!
    """
    reflectance_array, error_array, _wavelength_array = preprocess_input(
        reflectance, errors, _wavelengths
    )

    average_of_reflectance = np.mean(reflectance_array, axis=0)
    error_values = addition_in_quadrature(error_array)
    return reindex_or_mask(
        (average_of_reflectance, reflectance), (error_values, errors)
    )


def slope(
    reflectance: Union[pd.DataFrame, pd.Series, Sequence, np.ndarray],
    errors: Union[pd.DataFrame, pd.Series, Sequence, np.ndarray, None] = None,
    wavelengths: Union[Sequence] = None,
):
    """
    just a slope function. notionally, slope in reflectance-
    wavelength space between two filters. only looks at the first
    two 'rows' you pass it, should you pass it more.
    """
    assert wavelengths is not None, "wavelengths must be passed to slope()"
    assert None not in wavelengths, "wavelengths must be passed to slope()"
    reflectance_array, error_array, wavelength_array = preprocess_input(
        reflectance, errors, wavelengths
    )
    difference = reflectance_array[1] - reflectance_array[0]
    distance = wavelength_array[1] - wavelength_array[0]
    slope_value = difference / distance
    error_values = addition_in_quadrature(error_array)
    if error_values is not None:
        error_values /= distance
    return reindex_or_mask(
        (slope_value, reflectance), (error_values, errors)
    )


def band_depth(
    reflectance: Union[pd.DataFrame, pd.Series, Sequence, np.ndarray],
    errors: Union[pd.DataFrame, pd.Series, Sequence, np.ndarray, None] = None,
    wavelengths: Union[Sequence] = None,
):
    assert (
        wavelengths is not None
    ), "wavelengths must be passed to band_depth()"
    assert (
        None not in wavelengths
    ), "wavelengths must be passed to band_depth()"
    reflectance_array, error_array, wavelength_array = preprocess_input(
        reflectance, errors, wavelengths
    )
    assert (
        np.shape(reflectance_array)[0] == 3
    ), "band depth requires exactly 3 bands"
    assert (
        np.shape(wavelength_array)[0] == 3
    ), "band depth requires exactly 3 bands"
    # just for clarity
    wave_left = wavelength_array[0]
    wave_right = wavelength_array[1]
    wave_middle = wavelength_array[2]
    if len({wave_left, wave_right, wave_middle}) != 3:
        raise ValueError(
            "band depth between a wavelength and itself is undefined"
        )
    if not (
        max(wave_left, wave_right) > wave_middle > min(wave_left, wave_right)
    ):
        raise ValueError(
            "band depth can only be calculated at a band within the "
            "chosen range."
        )
    distance = wave_middle - wave_left
    slope_value = slope(
        reflectance_array[0:2], None, np.array([wave_left, wave_right])
    )[0]
    continuum_ref = reflectance_array[0] + slope_value * distance
    band_depth_value = 1 - reflectance_array[2] / continuum_ref
    error_values = addition_in_quadrature(error_array)
    if error_values is not None:
        error_values /= continuum_ref
    return reindex_or_mask(
        (band_depth_value, reflectance), (error_values, errors)
    )


# TODO, maybe: implement band_depth_min
#  nobody actually seems to want it

# TODO: is there any meaningful way to calculate error
#  for these min and max functions?


def band_min(
    reflectance: Union[pd.DataFrame, pd.Series, Sequence, np.ndarray],
    _errors: Union[pd.DataFrame, pd.Series, Sequence, np.ndarray, None] = None,
    wavelengths: Optional[Sequence] = None,
):
    assert wavelengths is not None, "wavelengths must be passed to band_min()"
    assert None not in wavelengths, "wavelengths must be passed to band_min()"
    reflectance_array, _error_array, wavelength_array = preprocess_input(
        reflectance, None, wavelengths
    )
    return reindex_or_mask(
        (
            wavelength_array[np.argmin(reflectance_array, axis=0)],
            reflectance,
        ),
        (None, None),
    )


def band_max(
    reflectance: Union[pd.DataFrame, pd.Series, Sequence, np.ndarray],
    _errors: Union[pd.DataFrame, pd.Series, Sequence, np.ndarray, None] = None,
    wavelengths: Optional[Sequence] = None,
):
    assert wavelengths is not None, "wavelengths must be passed to band_max()"
    assert None not in wavelengths, "wavelengths must be passed to band_max()"
    reflectance_array, _error_array, wavelength_array = preprocess_input(
        reflectance, None, wavelengths
    )
    return reindex_or_mask(
        (
            wavelength_array[np.argmax(reflectance_array, axis=0)],
            reflectance,
        ),
        (None, None),
    )


SPECTOP_NAMES = (
    "band_avg",
    "band_depth",
    "band_max",
    "band_min",
    "ratio",
    "slope",
)



from marslab.imgops import imgutils
import numpy as np


class TestApplyImageFilter:
    def test_apply_image_filter_1(self):
        out = imgutils.apply_image_filter(
            [[10, 10]],
            image_filter={"function": np.mean, "params": {"axis": 0}},
        )
        assert out[0] == 10
        assert out[1] == 10

    def test_apply_image_filter_2(self):
        out = imgutils.apply_image_filter([[10, 10]])
        assert out[0][0] == 10
        assert out[0][1] == 10


class TestCrop:
    def test_crop_1(self):
        img = np.ones((10, 10))
        out = imgutils.crop(img, bounds=(1, 1, 1, 1))
        assert out.shape[0] == 8
        assert out.shape[1] == 8



from collections.abc import (
    Sequence,
    Mapping,
)
from typing import Union

import numpy as np

# bayer pattern definitions
from marslab.imgops.imgutils import bilinear_interpolate_subgrid

RGGB_PATTERN = {
    "red": (0, 0),
    "green_1": (0, 1),
    "green_2": (1, 0),
    "blue": (1, 1),
}


def make_pattern_masks(
    array_shape: Sequence[int, int],
    bayer_pattern: Mapping[str, Sequence[int]],
    pattern_shape: Sequence[int, int],
) -> dict[str, tuple]:
    """
    given (y, x) array shape and a dict of tuples defining grid class names
    and positions,
    generate a dict of array pairs containing y, x coordinates for each
    named grid class in a m x n pattern beginning at the upper-left-hand corner
    of an array of shape shape and extending across the entirety of that array,


    supports only m x n patterns (like conventional bayer patterns); not a
    'generalized n-d discrete
    frequency class sorter' or whatever
    """

    y_coord, x_coord = np.meshgrid(
        np.arange(array_shape[0]), np.arange(array_shape[1])
    )
    masks = {}
    for name, position in bayer_pattern.items():
        y_slice = slice(position[0], None, pattern_shape[0])
        x_slice = slice(position[1], None, pattern_shape[1])
        masks[name] = (
            np.ravel(y_coord[y_slice, x_slice]),
            np.ravel(x_coord[y_slice, x_slice]),
        )
    return masks


def make_bayer(
    array_shape: Sequence[int, int],
    bayer_pattern: Mapping[str, Sequence[int]],
) -> dict[str, tuple]:
    """
    alias to make_pattern_masks
    """
    return make_pattern_masks(array_shape, bayer_pattern, (2, 2))


def debayer_upsample(
    image: np.ndarray,
    pixel: Union[str, Sequence[str]],
    pattern: Mapping[str, tuple] = None,
    masks: Mapping[str, tuple] = None,
    row_column: Mapping[str, tuple] = None,
    **_kwargs  # TODO: hacky!
) -> np.ndarray:
    """
    debayer and upsample an image, given a bayer pixel name or names
    and either a bayer pattern or explicitly precalculated sets of absolute (and
    optionally relative) coordinates for those pixels.
    averages arrays if more than one pixel name is given.
    TODO: the preamble to this may be excessively convoluted.
    """
    assert not (pattern is None and masks is None), (
        "debayer_upsample() must be passed either a bayer pattern or "
        "precalculated bayer masks."
    )
    if isinstance(pixel, str):
        pixel = [pixel]
    if masks is None:
        masks = make_bayer(image.shape, pattern)
    if row_column is None:
        row_column = {
            pixel: (np.unique(mask[0]), np.unique(mask[1]))
            for pixel, mask in masks.items()
        }
    upsampled_images = []
    for pix in pixel:
        mask = masks[pix]
        if row_column is None:
            rows, columns = np.unique(mask[0]), np.unique(mask[1])
        else:
            rows, columns = row_column[pix]
        subframe = image[mask].reshape(rows.size, columns.size, order="F")
        upsampled_images.append(
            bilinear_interpolate_subgrid(rows, columns, subframe, image.shape)
        )
    if len(upsampled_images) == 1:
        return upsampled_images[0]
    return np.mean(np.dstack(upsampled_images), axis=-1)


# TODO: extract methods, clean up
def mask_bayer_pixels(
    image: np.ndarray,
    pixel: Union[str, Sequence[str]],
    pattern: Mapping[str, tuple] = None,
    masks: Mapping[str, tuple] = None,
    default = 0,
    **_kwargs  # TODO: hacky!
) -> np.ndarray:
    """
    return a version of an image with non-matching bayer pixels set to default
    """
    assert not (pattern is None and masks is None), (
        "debayer_blank() must be passed either a bayer pattern or "
        "precalculated bayer masks."
    )
    masked = image.copy()
    if isinstance(pixel, str):
        pixel = [pixel]
    if masks is None:
        masks = make_bayer(image.shape, pattern)
    for pix in masks.keys():
        if pix in pixel:
            continue
        mask = masks[pix]
        masked[mask] = default
    return masked


from _operator import truediv, floordiv
from functools import reduce
from inspect import signature
from types import MappingProxyType
from typing import Collection

import numpy as np
import scipy.ndimage as ndi
from more_itertools import windowed
from skimage.measure import label

from marslab.imgops.imgutils import (
    ravel_valid,
    colorfill_maskedarray,
    cut_annulus,
    eightbit,
    normalize_range,
)


def threshold_mask(
    arrays: Collection[np.ndarray], percentiles=(1, 99), operator="or"
):
    masks = []
    if operator == "mean":
        arrays = [_get_flat_mean(arrays, None)]
    for array in arrays:
        low, high = np.percentile(ravel_valid(array), percentiles)
        masks.append(np.ma.masked_outside(array, low, high).mask)
    if operator == "mean":
        return masks[0]
    if operator == "or":
        return reduce(np.logical_or, masks)
    return reduce(np.logical_and, masks)


def ink(
    images,
    median=5,
    minimum=5,
    maximum=3,
    edge_thresholds=(60, 110),
    erosion=3,
    input_stretch=(10, 1)
):
    mean = maybe_filter(
        _get_flat_mean(images, None, normalize=input_stretch), median
    )
    edges = outline(mean, edge_thresholds, maximum, erosion=erosion)
    canvas = np.ones(images[0].shape)
    canvas[edges] = 0
    return maybe_filter(canvas, minimum, filt=ndi.minimum_filter)


def centile_threshold(array, percentile):
    return array > np.percentile(ravel_valid(array), percentile)


def pluck_label(array, connectivity=1, label_ix=1):
    mask = label(array, connectivity=connectivity) == label_ix
    return mask


def floodfill_mask(mask):
    import cv2

    cvmask = mask.astype("uint8")
    contours, hierarchy = cv2.findContours(
        cvmask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_TC89_KCOS
    )
    flood = cv2.fillPoly(cvmask, contours, 1)
    return flood.astype("bool")


def crop_bounds(mask):
    rows, cols = np.any(mask, axis=1), np.any(mask, axis=0)
    return mask[rows][:, cols], rows, cols


def find_widest_row(mask):
    cropped, rows, cols = crop_bounds(mask)
    contiguous = []
    for row in cropped:
        truthy = [
            a
            for a in np.split(row, np.nonzero(np.diff(row))[0] + 1)
            if a.any()
        ]
        contiguous.append(max([t.size for t in truthy]))
    contiguous = np.array(contiguous)
    return np.nonzero(rows)[0][
        np.nonzero(contiguous == contiguous.max())[0][-1]
    ]


def get_widest_section(row):
    split_indices = np.nonzero(np.diff(row))[0] + 1
    length_indices = {}
    for ix, section in enumerate(np.split(row, split_indices)):
        if section.any():
            length_indices[ix] = section
    widest = max([len(v) for v in length_indices.values()])
    widest_section = [
        k for k, v in length_indices.items() if len(v) == widest
    ][0]
    return tuple(windowed([0, *split_indices, len(row) - 1], 2))[
        widest_section
    ]


def test_axes(mask, shape):
    vertical, horizontal = shape
    if (vertical is None) and (horizontal is None):
        return True
    col_indices, row_ix = widest_part(mask)
    if vertical is not None:
        above = mask[:row_ix, col_indices[0] : col_indices[1]]
        try:
            if (np.nonzero(above)[0].size / above.size) < vertical:
                return False
        except ZeroDivisionError:
            pass
    if horizontal is not None:
        if ((col_indices[1] - col_indices[0]) / mask.shape[1]) < horizontal:
            return False
    return True


def test_coverage(mask, cutoff):
    if cutoff is None:
        return True
    mask_extent = mask[np.nonzero(mask)].size
    filled_extent = mask[np.nonzero(floodfill_mask(mask))].size
    if (mask_extent / filled_extent) < cutoff:
        return False
    return True


def test_extent(mask, cutoff):
    if cutoff is None:
        return True
    if (mask[np.nonzero(mask)].size / mask.size) < cutoff:
        return False
    return True


def _blank(shape):
    return np.full(shape, False)


def maybe_filter(array, size=None, footprint=None, filt=ndi.median_filter):
    if (size is None) and (footprint is None):
        return array
    if footprint is None:
        footprint = np.ones((size, size))
    sig = signature(filt)
    for kwarg in ("structure", "footprint"):
        if kwarg in sig.parameters.keys():
            break
    return filt(array, **{kwarg: footprint})


def widest_part(mask):
    widest_row_ix = find_widest_row(mask)
    widest_row = mask[widest_row_ix]
    w_col_indices = get_widest_section(widest_row)
    return w_col_indices, widest_row_ix


def check_mask_validity(mask, extent=None, coverage=None, v=None, h=None):
    for check, threshold in zip(
        (test_extent, test_coverage, test_axes), (extent, coverage, (v, h))
    ):
        if check(mask, threshold) is False:
            return False
    return True


def outline(array, edge_thresholds=(60, 100), maximum=5, erosion=3):
    import cv2

    if array.dtype != np.uint8:
        array = eightbit(array)
    edges = cv2.Canny(array, *edge_thresholds)
    edges = maybe_filter(edges, maximum, filt=ndi.maximum_filter)
    return maybe_filter(edges, erosion, filt=ndi.binary_erosion)


def _get_flat_mean(arrays, dilate_mean=None, normalize=None):
    if normalize is not None:
        arrays = [
            normalize_range(array, stretch=normalize) for array in arrays
        ]
    if all([isinstance(a, np.ma.MaskedArray) for a in arrays]):
        mean = _flat_masked_mean(arrays, dilate_mean)
    else:
        mean = np.ma.mean(np.ma.dstack(arrays), axis=-1)
    return mean


def clear_overhead_pixels(mask, row, start, stop):
    """warning: mutates mask inplace"""
    mask[:row, start:stop] = 1
    return mask


def trim_infiltrate_pixels(mask, row=0, trim_depth=10, depth_sigma=7):
    not_mask = ~floodfill_mask(mask).copy()
    not_mask[:row] = 0
    depth = ndi.gaussian_filter(np.cumsum(not_mask, axis=0), depth_sigma)
    depth_mask = depth <= trim_depth
    return np.logical_and(mask, depth_mask)


def refit_mask(
    mask, clear=True, trim=False, trim_depth=10, depth_sigma=7
):
    if (clear is False) and (trim is False):
        return mask
    w_col_indices, w_row_ix = widest_part(floodfill_mask(mask))
    if clear is True:
        mask = clear_overhead_pixels(
            mask, w_row_ix, w_col_indices[0], w_col_indices[-1]
        )
    if trim is True:
        mask = trim_infiltrate_pixels(
            mask, w_row_ix, trim_depth, depth_sigma
        )
    return pluck_label(mask)


def pick_valid_label(array, cutoffs, nanmask=None):
    labels = label(array)
    for label_ix in np.unique(labels[0][labels[0] >= 1]):
        mask = labels == label_ix
        if nanmask is not None:
            checkmask = np.logical_or(mask, nanmask)
        else:
            checkmask = mask
        if check_mask_validity(checkmask, **cutoffs) is True:
            return mask
    return None


def superpixel_edges(
    images,
    stretch=(10, 1),
    slic_kwargs=MappingProxyType(
        {"n_segments": 5, "sigma": 2, "compactness": 3}
    ),
):
    from skimage.segmentation import find_boundaries, slic

    cube = np.dstack(
        [normalize_range(image, stretch=stretch) for image in images]
    )
    return find_boundaries(slic(cube, **slic_kwargs))


def skymask(
    arrays,
    percentile=75,
    edge_params=MappingProxyType({"maximum": 5, "erosion": 3}),
    input_median=5,
    trace_maximum=5,
    cutoffs=MappingProxyType(
        {"extent": 0.05, "coverage": None, "v": 0.9, "h": None}
    ),
    input_mask_dilation=None,
    input_stretch=(10, 1),
    floodfill=True,
    trim_params=MappingProxyType({"trim": False}),
    clear=True,
    colorblock=False,
    respect_mask=False
):
    flat_mean = _get_flat_mean(
        arrays, input_mask_dilation, normalize=input_stretch
    )
    mean = maybe_filter(flat_mean, input_median)
    if respect_mask is True:
        mean[flat_mean.mask] = np.nan
    trace = ~outline(mean, **edge_params)
    if respect_mask is True:
        nanmask = maybe_filter(
            flat_mean.mask.copy(), input_mask_dilation, filt=ndi.maximum_filter
        )
    else:
        nanmask = None
    if percentile is not None:
        trace = np.logical_and(trace, centile_threshold(mean, percentile))
    if colorblock is True:
        trace[superpixel_edges(arrays)] = 0
    if trace_maximum is not None:
        trace = ndi.minimum_filter(trace, trace_maximum)
    if (mask := pick_valid_label(trace, cutoffs, nanmask)) is None:
        return _blank(mean.shape)
    mask = refit_mask(mask, clear, **trim_params)
    if floodfill is True:
        return floodfill_mask(mask)
    return mask


def _flat_masked_mean(arrays, dilate=None):
    if dilate is not None:
        mask = flatmask(arrays, dilate=dilate)
    else:
        mask = reduce(np.logical_or, [a.mask for a in arrays])
    mean = np.ma.MaskedArray(np.mean(np.dstack(arrays), axis=-1), mask=mask)
    return mean


def extract_masks(images, instructions=None):
    if instructions is None:
        return images, None, []
    passmasks, sendmasks = [], []
    if any((inst.get("colorfill") is not None for inst in instructions)):
        canvas = np.zeros(images[0].shape)
    for inst in instructions:
        do_pass, do_send = inst.get("pass", False), inst.get("send", True)
        if (do_pass is False) and (do_send is False):
            continue
        func, params = inst["function"], inst["params"]
        mask = func(images, **params)
        if do_pass is True:
            passmasks.append(mask)
        if do_send is True:
            if (color_kwargs := inst.get("colorfill")) is not None:
                sendmasks.append(
                    colorfill_maskedarray(
                        np.ma.masked_array(canvas, mask), **color_kwargs
                    )
                )
            else:
                sendmasks.append(mask)
    flattened_mask = (
        None if len(passmasks) == 0 else reduce(np.logical_or, passmasks)
    )
    return images, flattened_mask, sendmasks


def flatmask(images, dilate=None, sharp=False, square=False):
    try:
        mask = reduce(
            np.logical_or,
            [i.mask for i in images if isinstance(i, np.ma.MaskedArray)],
        )
    except TypeError as te:
        if "empty iterable" not in str(te):
            raise
        return np.full(images[0].shape, False)
    if dilate is not None:
        # noinspection PyTypeChecker
        return dilate_mask(mask, dilate, sharp, square)


def dilation_kernel(size=2, sharp=False, square=False):
    if size < 2:
        raise ValueError("Kernel size must be at least 2.")
    kernel = np.ones((size, size))
    if square is True:
        return kernel
    op = truediv if sharp is False else floordiv
    distance = op(size, 2)
    return np.ma.filled(cut_annulus(kernel, (0, distance)), 0)


def dilate_mask(image, size, sharp=False, square=False, copy=True):
    if not isinstance(image, np.ma.MaskedArray):
        return ndi.binary_dilation(
            image.astype(bool), structure=dilation_kernel(size, sharp, square)
        )
    if copy is True:
        image = image.copy()
    image.mask = ndi.binary_dilation(
        image.mask, structure=dilation_kernel(size, sharp, square)
    )
    return image



import numpy as np
from marslab.imgops.imgutils import normalize_range
from marslab.imgops.render import decorrelation_stretch
from marslab.tests.utilz.utilz import normal_array

rng = np.random.default_rng()


def test_decorrelation_stretch_1():
    for _ in range(100):
        channels = [normal_array() for _ in range(3)]
        stretched = decorrelation_stretch(channels, contrast_stretch=1)
        print(np.std(normalize_range(np.dstack(channels))), np.std(stretched))
        assert np.std(normalize_range(np.dstack(channels))) < np.std(stretched)

#
# @given(
#     channels=(
#         st.lists(
#             arrays(
#                 dtype=np.float32,
#                 shape=st.shared(
#                     array_shapes(
#                         min_dims=2, max_dims=2, min_side=2, max_side=20
#                     ),
#                     key="hi"
#                 ),
#                 elements={
#                     "allow_nan": False,
#                     "allow_infinity": False,
#                     "min_value": 0,
#                     "max_value": 256,
#                 },
#                 fill=st.nothing(),
#             ),
#             min_size=2,
#         )
#     )
# )
# def test_decorrelation_stretch_2(channels):
#     # single-valued arrays have undefined covariance
#     assume(all([not len(np.unique(channel)) == 1 for channel in channels]))
#     # ridiculously large inputs relative to dtype we expect to cause errors
#     assume(all([not np.isinf(np.std(channel)) for channel in channels]))

# def check_dcs_input(channels):
#     if len(channels) < 2:
#         raise ValueError(
#             "At least two channels must be passed to this function"
#         )
#     if not all_equal([channel.shape for channel in channels]):
#         raise ValueError(
#             "arrays passed to this function must have equal shapes"
#         )
#     if not all_equal([channel.dtype for channel in channels]):
#         raise ValueError(
#             "arrays passed to this function must have equal dtypes"
#         )
#     if (
#         channels[0].dtype.char
#         not in np.typecodes["AllInteger"] + np.typecodes["AllFloat"]
#     ):
#         raise ValueError(
#             "arrays passed to this function must be integer or float arrays"
#         )

#



import numpy as np

from marslab.tests.utilz.bandset_utilz import (
    make_random_bandset,
)


def test_random_bandset():
    bandset = make_random_bandset()
    bandset.load()
    assert len(bandset.raw.keys()) == len(bandset.metadata)
    assert np.all(bandset.get_band(1) == bandset.raw[1])
    bandset.debayer_if_required(1)
    bandset.bulk_debayer(list(bandset.metadata['BAND'].values))
    assert bandset.debayered == {}


import numpy as np

import marslab.imgops.imgutils
from hypothesis import given, strategies as st, settings as hset, Verbosity

from hypothesis.extra.numpy import (
    arrays,
    floating_dtypes,
    integer_dtypes,
    array_shapes,
)

# def positive_integer_arrays(**array_kwargs):
#     return arrays(
#         dtype=integer_dtypes(),
#         elements={"min_value": 0},
#         **array_kwargs
#     )
#
# rng = np.random.default_rng()
# @hset(verbosity=Verbosity.verbose)
# @given(
#     input_array=arrays(
#         dtype=floating_dtypes(), shape=array_shapes(max_dims=2)
#     ),
# )
# def test_fuzz_bilinear_interpolate_subgrid(
#     input_array
# ):
#     rows = rng.choice(np.arange(0, input_array.shape[1] * 2,
#     cols = rng.choice(np.arange(0, input_array.shape[0], num_cols))
#     output_shape = (input_array.shape[0]*2, input_array.shape[1]*2)
#     marslab.imgops.imgutils.bilinear_interpolate_subgrid(
#         rows=rows,
#         columns=cols,
#         input_array=input_array,
#         output_shape=output_shape,
#     )



"""
skeleton BandSet implementation for M3 images -- expects files that have been
processed by pdr.converter
"""

from functools import partial

import numpy as np
import pandas as pd

from marslab.compat.m3 import (
    parse_m3_fn,
    TARGET_WAVELENGTHS,
    GLOBAL_WAVELENGTHS,
)
from marslab.bandset.bandset import BandSet
from marslab.imgops.loaders import pdr_load


def setup_m3_bandset_metadata(file_path):
    fn_metadata = parse_m3_fn(str(file_path))
    metadata = pd.DataFrame()
    if fn_metadata["mode"] == "target":
        metadata["WAVELENGTH"] = TARGET_WAVELENGTHS
    elif fn_metadata["mode"] == "global":
        metadata["WAVELENGTH"] = GLOBAL_WAVELENGTHS
    else:
        raise ValueError("don't recognize this instrument mode")
    metadata["IX"] = np.array(range(len(metadata["WAVELENGTH"])))
    # 'channel' or 'channel #' in mission nomenclature
    metadata["BAND"] = metadata["IX"] + 1
    metadata['PATH'] = file_path
    return metadata


class M3BandSet(BandSet):
    def __init__(self, file_path, rois=None, threads=None):
        metadata = setup_m3_bandset_metadata(file_path)
        load_method = partial(pdr_load, preserve_constants=-999.0)
        super().__init__(
            metadata=metadata,
            load_method=load_method,
            rois=rois,
            threads=threads,
        )



from io import BytesIO

from PIL import Image
from hypothesis import given, strategies as st, reject, assume
from matplotlib.figure import Figure
import numpy as np

from marslab.imgops.imgutils import eightbit


positive_finite_floats = st.floats(
    min_value=0, allow_nan=False, allow_infinity=False
)

finite_floats = st.floats(allow_nan=False, allow_infinity=False)

finite_numbers = st.one_of(st.integers(), finite_floats)


def normal_array(loc=128, scale=16, size=(256, 256), dtype=np.int64):
    rng = np.random.default_rng()
    randarray = rng.normal(loc, scale, size)
    return randarray.astype(dtype)


def render_in_memory(fig: Figure) -> Image:
    outbuf = BytesIO()
    fig.savefig(outbuf)
    outbuf.seek(0)
    return Image.open(outbuf)



from setuptools import setup, find_packages

# this is provided as a convenience for automated install
# hooks. we do not recommend using this file to install
# marslab or its dependencies. please use conda along with
# the provided environment.yml file.
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="marslab",
    version="0.9.13",
    url="https://github.com/millionconcepts/marslab.git",
    author="Million Concepts",
    author_email="chase@millionconcepts.com",
    description="Utilities for working with observational data of Mars.",
    packages=find_packages(),
    python_requires=">=3.9",
    install_requires=[
        "numpy",
        "scipy",
        "dustgoggles",
        "fs",
        "pandas",
        "more-itertools",
        "pathos",
        "cytoolz",
    ],
    extras_require={
        "pdr_load": ["pdr"],
        "tests": ["pytest", "hypothesis"],
        "render": ["pillow", "matplotlib"],
        "notebooks": ["jupyter"],
        "regions": ["astropy"],
        "time": ["sympy", "python-dateutil"],
        "geom": ["pdr"],
        "strict_reshape": ["sympy"],
        "histograms": ["fast-histogram"],
        "masking": ["scikit-image", "opencv"]
    }
)



"""
geometry utilities based on metadata in object labels.
"""
from itertools import product
from functools import reduce
from operator import or_

from dustgoggles.func import is_it
from dustgoggles.structures import NestingDict, get_from
import numpy as np
import pandas as pd
import pdr


def get_geometry_value(entity_name, frame_name, axis_name, data: pdr.Data):
    return get_from(
        data.metaget(f"{frame_name}_DERIVED_GEOMETRY_PARMS"),
        (f"{entity_name}_{axis_name}", "value")
    )


def get_coordinate_system_properties(frame_name, data):
    system = data.metaget(f"{frame_name}_COORDINATE_SYSTEM")
    if system is None:
        return None
    return {
        "name": system.get("COORDINATE_SYSTEM_NAME"),
        "reference_frame": system.get("REFERENCE_COORD_SYSTEM_NAME"),
        "quaternion": np.array(system.get("ORIGIN_ROTATION_QUATERNION")),
        "offset": np.array(system.get("ORIGIN_OFFSET_VECTOR")),
        "orientation": system.get("POSITIVE_AZIMUTH_DIRECTION")
    }


def get_coordinate_systems(data: pdr.Data):
    frame_names = [
        k.replace("_COORDINATE_SYSTEM", "")
        for k in data.metadata.fieldcounts
        if k.endswith("_COORDINATE_SYSTEM")
    ]
    return {
        name: get_coordinate_system_properties(name, data)
        for name in frame_names
    }


def cart2sph(x0, y0, z0, unit: str = "degrees"):
    """
    convert cartesian to spherical coordinates. returns degrees by default;
    pass unit="radians" to return radians. if passed any arraylike objects,
    returns a DataFrame, otherwise, returns a tuple of values.

    caveats:
    1. this assumes a coordinate convention in which latitude runs from -90
        to 90 degrees.
    2. returns longitude in strictly positive coordinates.
    """
    radius = np.sqrt(x0 ** 2 + y0 ** 2 + z0 ** 2)
    if x0 != 0:
        longitude = np.arctan2(y0, x0)
    else:
        longitude = np.pi / 2
    longitude = longitude % (np.pi * 2)
    latitude = np.arcsin(z0 / np.sqrt(x0 ** 2 + y0 ** 2 + z0 ** 2))
    if unit == "degrees":
        latitude = np.degrees(latitude)
        longitude = np.degrees(longitude)
    if reduce(
        or_, map(is_it(pd.DataFrame, np.ndarray, pd.Series), [x0, y0, z0])
    ):
        return pd.DataFrame({"lat": latitude, "lon": longitude, "r": radius})
    return latitude, longitude, radius


def sph2cart(lat, lon, radius=1, unit: str = "degrees"):
    """
    convert spherical to cartesian coordinates. assumes input is in degrees
    by default; pass unit="radians" to specify input in radians. if passed any
    arraylike objects, returns a DataFrame, otherwise, returns a tuple of
    values.

    caveats:
    1. this assumes a coordinate convention in which latitude runs from -90
        to 90 degrees.
    """
    if unit == "degrees":
        lat = np.radians(lat)
        lon = np.radians(lon)
    x0 = radius * np.cos(lat) * np.cos(lon)
    y0 = radius * np.cos(lat) * np.sin(lon)
    z0 = radius * np.sin(lat)

    if reduce(
        or_,
        map(is_it(pd.DataFrame, np.ndarray, pd.Series), [lat, lon, radius]),
    ):
        return pd.DataFrame({"x": x0, "y": y0, "z": z0})
    return x0, y0, z0


def quaternion_multiplication(q1, q2):
    s1, v1 = q1[0], q1[1:]
    s2, v2 = q2[0], q2[1:]
    scalar = s1 * s2 - np.dot(v1, v2)
    vector = (s1 * v2 + s2 * v1 + np.cross(v1, v2))
    return np.array([scalar, *vector])


def invert_quaternion(quaternion):
    return np.array([quaternion[0]] + [-1 * q for q in quaternion[1:]])


def rotate_unit_vector(alt, az, rotation_quaternion, orientation="clockwise"):
    """
    assumptions:
    1. left-handed coordinate system,
    2. scalar-first quaternion representation.
    3. units are degrees.
    """
    source_cartesian = np.array(sph2cart(alt, az))
    if orientation == "clockwise":
        source_cartesian *= np.array([-1, -1, 1])
    zero_quaternion = np.array([0, *source_cartesian])
    inverse_rotation = invert_quaternion(rotation_quaternion)
    q_times_0_v = quaternion_multiplication(
        rotation_quaternion, zero_quaternion
    )
    v_prime = quaternion_multiplication(q_times_0_v, inverse_rotation)
    assert np.isclose(v_prime[0], 0)
    target_cartesian = v_prime[1:]
    if orientation == "clockwise":
        target_cartesian *= np.array([-1, -1, 1])
    return np.array(cart2sph(*target_cartesian))


def get_coordinates(data: pdr.Data):
    # TODO: maybe expand this
    axes = ("AZIMUTH", "ELEVATION")
    entities = set()
    for axis in axes:
        entities.update({
            k.replace(f"_{axis}", "")
            for k in data.metadata.fieldcounts
            if k.endswith(f"_{axis}")
        })
    coordinates = NestingDict()
    parms = {
        k for k in data.metadata.fieldcounts
        if k.endswith("_DERIVED_GEOMETRY_PARMS")
    }
    for system, axis, entity in product(parms, axes, entities):
        block = data.metaget(system)
        if block is None:
            continue
        record = block.get(f"{entity}_{axis}")
        if record is not None:
            coordinates[
                system.replace("_DERIVED_GEOMETRY_PARMS", "")
            ][entity][axis] = record['value']
    return coordinates.todict()


def transform_angle(source_frame, target_frame, entity, data):
    coordinates = get_coordinates(data)
    systems = get_coordinate_systems(data)
    source_info = systems.get(source_frame)
    target_info = systems.get(target_frame)
    if source_info is None:
        if not target_info["reference_frame"].startswith(source_frame):
            raise ValueError("can't directly convert between these frames.")
        quaternion = invert_quaternion(target_info["quaternion"])
    else:
        quaternion = source_info["quaternion"]
    coord = coordinates[source_frame][entity]
    if target_info is not None:
        orientation = target_info['orientation'].lower()
    else:
        orientation = source_info['orientation'].lower()
    return rotate_unit_vector(
        coord['ELEVATION'], coord['AZIMUTH'], quaternion, orientation
    )
